#!/usr/bin/env python3
"""
Utility for PDB REST API.
https://www.rcsb.org/docs/programmatic-access/web-services-overview
https://data.rcsb.org/redoc/
"""
###
import sys,os,re,json,argparse,time,logging
#
from .. import pdb as pdb_utils
#
##############################################################################
if __name__=='__main__':
  epilog="""
Example entry IDs: 3ERT, 3TTC
Example chemical IDs: CFF
"""
  parser = argparse.ArgumentParser(description='PDB REST API client', epilog=epilog)
  ops = ['list_entrys', 
    'get_entrys', 'get_chemicals', 
    ]
  parser.add_argument("op", choices=ops, help='OPERATION')
  parser.add_argument("--ids", dest="ids", help="PDB entry IDs, comma-separated")
  parser.add_argument("--i", dest="ifile", help="input file, PDB entry IDs")
  parser.add_argument("--druglike", action="store_true", help="druglike chemicals only (organic; !polymer; !monoatomic)")
  parser.add_argument("--o", dest="ofile", help="output file (TSV)")
  parser.add_argument("--api_host", default=pdb_utils.API_HOST)
  parser.add_argument("--api_base_path", default=pdb_utils.API_BASE_PATH)
  parser.add_argument("-v", "--verbose", action="count", default=0)
  args = parser.parse_args()

  logging.basicConfig(format='%(levelname)s:%(message)s', level=(logging.DEBUG if args.verbose>1 else logging.INFO))

  base_url = f"https://{args.api_host}{args.api_base_path}"

  ids=[]
  if args.ifile:
    fin = open(args.ifile)
    while True:
      line = fin.readline()
      if not line: break
      ids.append(line.strip())
  elif args.ids:
    ids = re.split('[, ]+', args.ids.strip())

  fout = open(args.ofile, "w+") if args.ofile else sys.stdout

  t0=time.time()

  if args.op == "get_entrys":
    if not ids: parser.error('ID[s] required.')
    pdb_utils.GetEntrys(ids, base_url, fout)

  elif args.op == "get_chemicals":
    if not ids: parser.error('ID[s] required.')
    pdb_utils.GetChemicals(ids, args.druglike, base_url, fout)

  elif args.op == "list_entrys":
    pdb_utils.ListEntrys(base_url, fout)

  else:
    parser.error(f"Invalid operation: {args.op}")

  logging.info(f"Elapsed time: {time.strftime('%Hh:%Mm:%Ss', time.gmtime(time.time()-t0))}")



#!/usr/bin/env python3
"""
See: https://smart-api.info/ui/55be2831c69b17f6bc975ddb58cabb5e
"""
###
import sys,os,re,argparse,time,json,logging
import pandas as pd
#
from .. import ubkg
#
##############################################################################
if __name__=='__main__':
  epilog='''\
The Data Distillery Knowledge Graph (DDKG) is an extension and instance of the Unified Biomedical Knowledge Graph (UBKG), developed by the University of Pittsburgh, Children's Hospital of Philadelphia, and the Common Fund Data Ecosystem (CFDE) Data Distillery Partnership Project.
'''
  parser = argparse.ArgumentParser(description='UBKG REST API client', epilog=epilog)
  ops = ['search',
          'info',
          'list_relationship_types',
          'list_node_types',
          'list_node_type_counts',
          'list_property_types',
          'list_sabs',
          'list_semantic_types',
          'get_concept2codes',
          'get_concept2concepts',
          'get_concept2definitions',
          'get_concept2paths',
          'get_concept2trees',
          'get_term2concepts',
          ]
  parser.add_argument("op", choices=ops, help='OPERATION')
  parser.add_argument("--o", dest="ofile", help="output (TSV)")
  parser.add_argument("--i", dest="ifile", help="UMLS CUI ID file")
  parser.add_argument("--ids", help="UMLS CUI IDs, comma-separated")
  parser.add_argument("--term", help="UMLS term, e.g. 'Asthma'")
  parser.add_argument("--sab", default="SNOWMEDCT_US", help="Standard abbreviation type")
  parser.add_argument("--relationship", default="isa", help="Relationship type")
  parser.add_argument("--mindepth", type=int, default=1, help="min path depth")
  parser.add_argument("--maxdepth", type=int, default=3, help="max path depth")
  parser.add_argument("--nmax", type=int, default=None, help="max records")
  parser.add_argument("--skip", type=int, default=0, help="skip 1st SKIP queries")
  parser.add_argument("--api_host", default=ubkg.API_HOST)
  parser.add_argument("--api_base_path", default=ubkg.API_BASE_PATH)
  parser.add_argument("--api_key")
  parser.add_argument("-v", "--verbose", default=0, action="count")
  args = parser.parse_args()

  logging.basicConfig(format='%(levelname)s:%(message)s', level=(logging.DEBUG if args.verbose>1 else logging.INFO))

  BASE_URL='https://'+args.api_host+args.api_base_path

  fout = open(args.ofile,"w") if args.ofile else sys.stdout

  t0=time.time()

  ids=[];
  if args.ifile:
    with open(args.ifile) as fin:
      while True:
        line = fin.readline()
        if not line: break
        if line.rstrip(): ids.append(line.rstrip())
  elif args.ids:
    ids = re.split('[, ]+', args.ids.strip())

  if args.op == "info":
    ubkg.Info(args.api_key, BASE_URL, fout)

  elif args.op == "list_relationship_types":
    ubkg.ListRelationshipTypes(args.api_key, BASE_URL, fout)

  elif args.op == "list_node_types":
    ubkg.ListNodeTypes(args.api_key, BASE_URL, fout)

  elif args.op == "list_node_type_counts":
    ubkg.ListNodeTypeCounts(args.api_key, BASE_URL, fout)

  elif args.op == "list_property_types":
    ubkg.ListPropertyTypes(args.api_key, BASE_URL, fout)

  elif args.op == "list_semantic_types":
    ubkg.ListSemanticTypes(args.api_key, BASE_URL, fout)

  elif args.op == "list_sabs":
    ubkg.ListSABs(args.api_key, BASE_URL, fout)

  elif args.op == "get_concept2codes":
    ubkg.GetConcept2Codes(ids, args.api_key, BASE_URL, fout)

  elif args.op == "get_concept2concepts":
    ubkg.GetConcept2Concepts(ids, args.api_key, BASE_URL, fout)

  elif args.op == "get_concept2definitions":
    ubkg.GetConcept2Definitions(ids, args.api_key, BASE_URL, fout)

  elif args.op == "get_concept2paths":
    if not args.sab: parser.error(f"--sab required for operation: {args.op}")
    ubkg.GetConcept2Paths(ids, args.sab, args.relationship, args.mindepth, args.maxdepth, args.api_key, BASE_URL, fout)

  elif args.op == "get_concept2trees":
    if not args.sab: parser.error(f"--sab required for operation: {args.op}")
    ubkg.GetConcept2Trees(ids, args.sab, args.relationship, args.mindepth, args.maxdepth, args.api_key, BASE_URL, fout)

  elif args.op == "get_term2concepts":
    ubkg.GetTerm2Concepts(args.term, args.api_key, BASE_URL, fout)

  else:
    parser.error(f"Invalid operation: {args.op}")

  logging.info(('elapsed time: %s'%(time.strftime('%Hh:%Mm:%Ss', time.gmtime(time.time()-t0)))))



#!/usr/bin/env python3
'''
Commonly used functions for database client applications.
'''
import sys,os,logging,urllib.parse
import sqlalchemy

##############################################################################
def PostgreSqlConnect(dbhost, dbport, dbname, dbusr, dbpw):
  try:
    import psycopg2
    engine = sqlalchemy.create_engine(f"postgresql+psycopg2://{dbusr}:{urllib.parse.quote_plus(dbpw)}@{dbhost}:{dbport}/{dbname}")
  except Exception as e:
    logging.info(f"{e}")
    logging.error("Failed to connect.")
    return None
  return engine.connect()

##############################################################################
def MySqlConnect(dbhost, dbport, dbname, dbusr, dbpw):
  try:
    # https://pypi.org/project/mysql-connector-python/
    # pip install mysql-connector-python
    import mysql.connector
    engine = sqlalchemy.create_engine(f"mysql+mysqlconnector://{dbusr}:{urllib.parse.quote_plus(dbpw)}@{dbhost}:{dbport}/{dbname}")
  except Exception as e:
    logging.info(f"{e}")
    try:
      # https://pypi.org/project/PyMySQL/
      # pip install PyMySQL
      import pymysql
      engine = sqlalchemy.create_engine(f"mysql+pymysql://{dbusr}:{urllib.parse.quote_plus(dbpw)}@{dbhost}:{dbport}/{dbname}")
    except Exception as e:
      logging.info(f"{e}")
      try:
        # https://mysqlclient.readthedocs.io/
        # pip install mysqlclient
        import MySQLdb
        engine = sqlalchemy.create_engine(f"mysql+mysqldb://{dbusr}:{urllib.parse.quote_plus(dbpw)}@{dbhost}:{dbport}/{dbname}")
      except Exception as e:
        logging.info(f"{e}")
        logging.error("Failed to connect.")
        return None
  return engine.connect()



#!/usr/bin/env python3
"""
Utility app for the PubMed REST-ful-ish webservices API.
"""
###
import sys,os,re,argparse,time,logging
#
from .. import pubmed
#
##############################################################################
if __name__=='__main__':
  OPS = [
        "get_esummary",
        "get_record",
        ]
  parser = argparse.ArgumentParser(description="PubMed webservices client")
  parser.add_argument("op", choices=OPS, help="OPERATION")
  parser.add_argument("--i", dest="ifile", help="input IDs file (PMIDs)")
  parser.add_argument("--ids", help="input IDs (PMIDs) (comma-separated)")
  parser.add_argument("--o", dest="ofile", help="output (usually TSV)")
  parser.add_argument("--api_host", default=pubmed.API_HOST)
  parser.add_argument("--api_base_path", default=pubmed.API_BASE_PATH)
  parser.add_argument("--skip", type=int, default=0)
  parser.add_argument("--nmax", type=int, default=None)
  parser.add_argument("-q", "--quiet", action="store_true", help="Suppress progress notification.")
  parser.add_argument("-v", "--verbose", default=0, action="count")
  args = parser.parse_args()

  logging.basicConfig(format='%(levelname)s:%(message)s', level=(logging.DEBUG if args.verbose>0 else logging.ERROR if args.quiet else 15))

  base_url = f"https://{args.api_host}{args.api_base_path}"

  fout = open(args.ofile, "w") if args.ofile else sys.stdout

  ids=[]
  if args.ifile:
    fin = open(args.ifile)
    while True:
      line = fin.readline()
      if not line: break
      ids.append(line.rstrip())
    fin.close()
  elif args.ids:
    ids = re.split(r'[,\s]+', args.ids)
  logging.info(f"Input IDs: {len(ids)}")

  t0=time.time()

  if args.op == 'get_esummary':
    pubmed.GetESummary(ids, args.skip, args.nmax, base_url, fout)

  elif args.op == 'get_record':
    pubmed.GetRecord(ids, args.skip, args.nmax, base_url, fout)

  else:
    parser.error(f"Invalid operation: {args.op}")

  logging.info(('elapsed time: %s'%(time.strftime('%Hh:%Mm:%Ss',time.gmtime(time.time()-t0)))))




#!/usr/bin/env python3
"""
Bioregistry.io
"""
import sys,os,re,time,json,requests,logging
import pandas as pd
#
API_HOST='bioregistry.io'
API_BASE_PATH='/api'
API_BASE_URL='https://'+API_HOST+API_BASE_PATH
#
##############################################################################
def ListEntities(etype, base_url=API_BASE_URL, fout=None):
  "Entities: contributors|contexts|collections|registry|metaregistry"
  tags=None; df=None; n_out=0;
  response = requests.get(f"{base_url}/{etype}")
  logging.debug(json.dumps(response.json(), indent=2))
  results = response.json()
  for id_this,thing in results.items():
    if not tags:
      tags = list(thing.keys())
      for tag in tags[:]:
        if type(thing[tag]) in (list, dict):
          logging.info(f"Ignoring field: {tag}")
          tags.remove(tag)
    data = {"id":[id_this]}
    data.update({tag:[thing[tag] if tag in thing else None] for tag in tags})
    df_this = pd.DataFrame(data)
    if fout is None: df = pd.concat([df, df_this])
    else: df_this.to_csv(fout, sep="\t", index=False, header=bool(n_out==0))
    n_out += df_this.shape[0]
  logging.info(f"n_out ({etype}): {n_out}")
  return df

#############################################################################
def GetReference(ids, prefix, base_url=API_BASE_URL, fout=None):
  df=None; n_out=0;
  for id_this in ids:
    response = requests.get(f"{base_url}/reference/{prefix}:{id_this}")
    logging.debug(json.dumps(response.json(), indent=2))
    result = response.json()
    providers = result["providers"] if "providers" in result else []
    for provider,url_this in providers.items():
      df_this = pd.DataFrame({"prefix":[prefix], "id":[id_this], "provider_name":[provider], "provider_url":[url_this]})
      if fout is None: df = pd.concat([df, df_this])
      else: df_this.to_csv(fout, sep="\t", index=False, header=bool(n_out==0))
      n_out += df_this.shape[0]
  logging.info(f"n_out: {n_out}")
  return df

##############################################################################



#!/usr/bin/env python3
##############################################################################
### Utility functions for OpenPHACTS REST API.
###
### formats: JSON, XML, TSV
###
### Ref: https://dev.openphacts.org/docs
###      https://dev.openphacts.org/docs/1.5
###      https://dev.openphacts.org/docs/2.0
###
### Top level resources:
###    /assay
###    /compound
###    /disease
###    /getConceptDescription
###    /mapUri
###    /pathway
###    /pharmacology
###    /search
###    /sources
###    /structure
###    /target
###    /tissue
###    /tree
##############################################################################
### From https://dev.openphacts.org/docs/develop
#  ?oidd_assay_uri a ?bao_type ;
#     skos:exactMatch ?pubchem_assay ;
#     void:inDataset <http://rdf.ncats.nih.gov/opddr> .
###
#  /assay?uri=http%3A%2F%2Fopeninnovation.lilly.com%2Fbioassay%2329"
##############################################################################
import sys,os,re,time,logging
import requests,json,urllib.parse
import pandas as pd
#
API_HOST='beta.openphacts.org'
API_BASE_PATH='/2.1'
#
##############################################################################
def ShowCounts(user_id, user_key, base_url):
  countpaths=[
	'/assay/members/count',
	'/pathways/count',
	'/pharmacology/filters/activities/count'
	]
  for countpath in countpaths:
    url = f"{base_url}{countpath}?app_id={user_id}&app_key={user_key}"
    response = requests.get(url)
    logging.debug(json.dumps(response.json(), sort_keys=True, indent=2))
    logging.info(f"{countpath}: {response.content}")
#
##############################################################################
def ListPathways(user_id, user_key, base_url, fout):
  url = base_url+'/pathways?'
  url+=('&app_id=%s&app_key=%s'%(user_id,user_key))
  url+=('&_format=tsv')
  response = requests.get(url)
  logging.debug(json.dumps(response.json(), sort_keys=True, indent=2))
  lines = response.content.splitlines()
  tags = re.split('\t', lines[0])
  tags = [csv_utils.ToStringForCSV(tag) for tag in tags]
  fout.write((','.join(tags))+'\n')
  n_pth=0; n_out=0;
  for line in lines[1:]:
    vals = re.split('\t', line)
    vals = [csv_utils.ToStringForCSV(val) for val in vals]
    fout.write((','.join(vals))+'\n')
    n_pth+=1
    n_out+=1
  logging.info('n_col: %d'%len(tags))
  logging.info('n_pth: %d'%n_pth)
  logging.info('n_out: %d'%n_out)

#############################################################################
def ListSources(user_id, user_key, base_url,fout):
  url=base_url+'/sources?'
  url+=('&app_id=%s&app_key=%s'%(user_id,user_key))
  url+=('&_format=json')
  response = requests.get(url)
  result = response.json()

  if type(result) is not dict:
    return

  logging.debug(json.dumps(result, indent=2))

  SRC_TAGS=['title','description','homepage','landingPage']
  SRC_SUB_TAGS=['url','title','type','description','landingPage','dataDump']
  fout.write('i,'+(','.join(SRC_TAGS+['sub:'+tag for tag in SRC_SUB_TAGS]))+'\n')

  result = result['result'] if result.has_key('result') else {}
  primaryTopic = result['primaryTopic'] if result.has_key('primaryTopic') else {}
  dbs = primaryTopic['subset'] if primaryTopic.has_key('subset') else []

  logging.debug('len(dbs) = %d '%len(dbs))

  n_db=0; n_out=0; n_err=0; n_subdb=0;
  for db in dbs:
    n_db+=1
    title = db['title'] if db.has_key('title') else ''
    logging.debug('db title = "%s"'%title)
    vals=[]
    for tag in SRC_TAGS:
      vals.append(csv_utils.ToStringForCSV(db[tag]) if db.has_key(tag) else '')

    subdbs  = db['subset'] if db.has_key('subset') else []

    if type(subdbs) is not list:
      logging.error('subdbs="%s"'%str(subdbs))
      n_err+=1
      continue

    n_subdb_this=0;
    for subdb in subdbs:
      n_subdb_this+=1
      if type(subdb) is not dict:
        logging.error('subdb="%s"'%str(subdb))
        fout.write(('%d.%d,'%(n_db,n_subdb_this))+(','.join(vals+['' for tag in SRC_SUB_TAGS[1:]]))+'\n')
        n_out+=1
        continue

      title = subdb['title'] if subdb.has_key('title') else ''
      logging.debug('subdb title = "%s"'%title)

      vals_this = vals[:]
      for tag in SRC_SUB_TAGS:
        vals_this.append(csv_utils.ToStringForCSV(subdb[tag]) if subdb.has_key(tag) else '')

      fout.write(('%d.%d,'%(n_db,n_subdb_this))+(','.join(vals_this))+'\n')
      n_out+=1

    if not subdbs and title:
      fout.write(('%d,'%n_db)+(','.join(vals+['' for tag in SRC_SUB_TAGS]))+'\n')
      n_out+=1
    n_subdb+=n_subdb_this

  logging.info( 'n_db: %d'%n_db)
  logging.info( 'n_subdb: %d'%n_subdb)
  logging.info( 'n_out: %d'%n_out)
  logging.info( 'n_err: %d'%n_err)

##############################################################################
def ListTargetTypes(user_id, user_key, base_url,fout):
  url=base_url+'/target/types?'
  url+=('&app_id=%s&app_key=%s'%(user_id,user_key))
  url+=('&_format=json')
  response = requests.get(url)
  result = response.json()

  if type(result) is not dict:
    logging.error('result not dict')
    return

  result = result['result']
  if type(result) is not dict:
    logging.error('result not dict')
    return

  pt = result['primaryTopic']
  if type(pt) is not dict:
    logging.error('primaryTopic not dict')
    return

  ttyps = result['result']['primaryTopic']['hasTargetType']

  n_typ=0; n_out=0; tags=[];
  for ttyp in  ttyps:
    if n_typ==0 or not tags:
      tags = ttyp.keys()
      fout.write(','.join([csv_utils.ToStringForCSV(tag) for tag in tags])+'\n')
    vals=[]
    for tag in tags:
      vals.append(ttyp[tag] if ttyp.has_key(tag) else '')

    fout.write(','.join([csv_utils.ToStringForCSV(val) for val in vals])+'\n')
    n_typ+=1
    n_out+=1

  logging.info( 'n_col: %d'%len(tags))
  logging.info( 'n_typ: %d'%n_typ)
  logging.info( 'n_out: %d'%n_out)

##############################################################################
def ListTargets(uri_qry,user_id, user_key, base_url,fout):
  '''URL must be from supported hierarchy such as Chembl.  http://purl.uniprot.org/enzyme/6.2.-.-'''
  url=base_url+'/target/members/pages?uri=%s'%urllib2.quote(uri_qry).replace('/','%2F')
  url+=('&app_id=%s&app_key=%s'%(user_id,user_key))
  url+=('&_format=json')
  pagesize=10;
  url+=('&_pageSize=%d'%pagesize)

  page=0;
  n_tgt=0; n_out=0; tags=[];

  while True:
    page+=1
    url_this = url+('&_page=%d'%page)
    response = requests.get(url_this)
    result = response.json()

    if not result: break

    if type(result) is not dict:
      logging.error('result not dict')
      break

    result = result['result']
    if type(result) is not dict:
      logging.error('result not dict')
      break

    tgts = result['result']['items']

    for tgt in  tgts:
      if n_tgt==0 or not tags:
        tags = tgt.keys()
        fout.write(','.join([csv_utils.ToStringForCSV(tag) for tag in tags])+'\n')
      vals=[]
      for tag in tags:
        vals.append(tgt[tag] if tgt.has_key(tag) else '')

      fout.write(','.join([csv_utils.ToStringForCSV(val) for val in vals])+'\n')
      n_tgt+=1
      n_out+=1

  logging.info( 'n_col: %d'%len(tags))
  logging.info( 'n_tgt: %d'%n_tgt)
  logging.info( 'n_out: %d'%n_out)

##############################################################################
def GetAssay(uri_qry, user_id, user_key, base_url,fout):
  url=base_url+'/assay?uri=%s'%urllib.parse.quote(uri_qry).replace('/','%2F')
  url+=('&app_id=%s&app_key=%s'%(user_id,user_key))
  url+=('&_format=tsv')
  response = requests.get(url)
  result = response.content

  lines = result.splitlines()
  tags = re.split('\t', lines[0])
  tags = [csv_utils.ToStringForCSV(tag) for tag in tags]
  fout.write((','.join(tags))+'\n')
  n_ass=0; n_out=0;
  for line in lines[1:]:
    vals = re.split('\t', line)
    vals = [csv_utils.ToStringForCSV(val) for val in vals]
    fout.write((','.join(vals))+'\n')
    n_ass+=1
    n_out+=1

  logging.info( 'n_col: %d'%len(tags))
  logging.info( 'n_ass: %d'%n_ass)
  logging.info( 'n_out: %d'%n_out)

#############################################################################
def GetInstance(resource, uri_qry, user_id, user_key, base_url, fout):
  url=base_url+'/%s?uri=%s'%(resource,urllib2.quote(uri_qry).replace('/','%2F'))
  url+=('&app_id=%s&app_key=%s'%(user_id,user_key))
  url+=('&_format=tsv')
  response = requests.get(url)
  result = response.content

  lines = result.splitlines()
  tags = re.split('\t', lines[0])
  tags = [csv_utils.ToStringForCSV(tag) for tag in tags]
  fout.write((','.join(tags))+'\n')
  n=0; n_out=0;
  for line in lines[1:]:
    n+=1
    vals = re.split('\t', line)
    vals = [csv_utils.ToStringForCSV(val) for val in vals]
    fout.write((','.join(vals))+'\n')
    n_out+=1

  logging.info( 'n_col: %d'%len(tags))
  logging.info( 'n: %d'%n)
  logging.info( 'n_out: %d'%n_out)

#############################################################################



#!/usr/bin/env python3
#############################################################################
### https://maayanlab.cloud/sigcom-lincs/#/API
#############################################################################
import sys,os,re,json,requests,tqdm,logging
import pandas as pd
import urllib.parse
#
API_HOST="maayanlab.cloud"
API_BASE_PATH="/sigcom-lincs/metadata-api"
BASE_URL='https://'+API_HOST+API_BASE_PATH
#
#############################################################################
def GetResources(ids, base_url=BASE_URL, fout=None):
  tags=None; df=None;
  url_base = (base_url+'/resources')
  for id_this in ids:
    url = f"{url_base}/{urllib.parse.quote(id_this)}"
    response = requests.get(url)
    rval = response.json()
    logging.debug(json.dumps(rval, indent=2))
    if not tags:
      tags = [tag for tag in rval.keys() if type(rval[tag]) not in (list, dict)]
    res = rval
    df = pd.concat([df, pd.DataFrame({tags[j]:[res[tags[j]]] for j in range(len(tags))})])
  if fout: df.to_csv(fout, sep="\t", index=False)
  logging.info(f"IDs: {len(ids)}")
  return df

#############################################################################



#!/usr/bin/env python3
"""
ClinicalTrials.gov
https://clinicaltrials.gov/api/gui/ref/api_urls (to be retired June 2024)
API v2.0:
https://clinicaltrials.gov/data-api/about-api
https://clinicaltrials.gov/data-api/api
https://clinicaltrials.gov/find-studies/constructing-complex-search-queries
"""
import sys,os,re,time,json,logging,requests,tqdm
import urllib,urllib.parse
import pandas as pd
#
API_HOST='clinicaltrials.gov'
API_BASE_PATH='/api/v2'
API_BASE_URL='https://'+API_HOST+API_BASE_PATH
API_PAGE_SIZE=50
#
##############################################################################
def Version(base_url=API_BASE_URL, fout=None):
  response = requests.get(f"{base_url}/version")
  result = response.json()
  df = pd.DataFrame(result, index=[0])
  if fout is not None: df.to_csv(fout, sep="\t", index=False)
  return df

##############################################################################
def ListStudyFields(base_url=API_BASE_URL, fout=None):
  response = requests.get(f"{base_url}/studies/metadata")
  logging.debug(response.text)
  result = response.json()
  df = pd.DataFrame(result)
  if fout is not None: df.to_csv(fout, sep="\t", index=False)
  logging.info(f"n_out: {df.shape[0]}")
  return df

##############################################################################
def ListSearchAreas(base_url=API_BASE_URL, fout=None):
  n_out=0; df=None;
  response = requests.get(f"{base_url}/studies/search-areas")
  logging.debug(response.text)
  result = response.json()
  areas = result[0]["areas"]
  for area in areas:
    name = area["name"]
    param = area["param"] if "param" in area else None
    parts = [part["pieces"][0] for part in area["parts"]]
    df_this = pd.DataFrame({
	"name":[name],
	"param":[param],
	"parts":[";".join(parts)],
	})
    if fout is not None: df_this.to_csv(fout, sep="\t", index=False, header=bool(n_out==0))
    else: pd.concat([df, df_this])
    n_out+=1
  logging.info(f"n_out: {n_out}")
  return df

##############################################################################
def ListEnums(base_url=API_BASE_URL, fout=None):
  n_out=0; df=None;
  response = requests.get(f"{base_url}/studies/enums")
  logging.debug(response.text)
  result = response.json()
  enums = result
  for enum in enums:
    values = [value["value"] for value in enum["values"]]
    legacyValues = [value["legacyValue"] for value in enum["values"]]
    #exceptions = enum["exceptions"] if "exceptions" in enum else None #hash
    df_this = pd.DataFrame({
	"type":[enum["type"] for value in values],
	"value":values,
	"legacyValue":legacyValues,
	})
    if fout is not None: df_this.to_csv(fout, sep="\t", index=False, header=bool(n_out==0))
    else: pd.concat([df, df_this])
    n_out+=df_this.shape[0]
  logging.info(f"n_out: {n_out}")
  return df

##############################################################################
def SearchStudies(query_cond, query_term, base_url=API_BASE_URL, fout=None):
  study_tags=None; n_out=0; df=None; nextPageToken=None; totalCount=None; tq=None;
  url_base = f"{base_url}/studies?pageSize={API_PAGE_SIZE}"
  if query_cond: url_base += f"&query.cond={urllib.parse.quote(query_cond)}"
  if query_term: url_base += f"&query.term={urllib.parse.quote(query_term)}"
  while True:
    url = url_base
    if not totalCount: url += "&countTotal=true"
    if nextPageToken: url += f"&pageToken={nextPageToken}"
    response = requests.get(url)
    result = response.json()
    if not totalCount: totalCount = result["totalCount"]
    logging.debug(f"totalCount: {totalCount}")
    if not tq: tq = tqdm.tqdm(total=totalCount)
    studies = result["studies"]
    for study in studies:
      logging.debug(json.dumps(study, indent=2))
      tq.update()
      df_this = Study2DataFrame(study)
      if fout is None: df = pd.concat([df, df_this])
      else: df_this.to_csv(fout, sep="\t", index=False, header=bool(n_out==0))
      n_out += df_this.shape[0]
    nextPageToken = result["nextPageToken"] if "nextPageToken" in result else None
    if not nextPageToken: break
  if tq is not None: tq.close()
  logging.info(f"n_out: {n_out}")
  return df

##############################################################################
def GetStudies(ids, base_url=API_BASE_URL, fout=None):
  n_out=0; df=None; tq=None;
  for id_this in ids:
    if not tq: tq = tqdm.tqdm(total=len(ids))
    url = f"{base_url}/studies/{id_this}"
    response = requests.get(url)
    result = response.json()
    study = result
    tq.update()
    logging.debug(json.dumps(study, indent=2))
    df_this = Study2DataFrame(study)
    if fout is None: df = pd.concat([df, df_this])
    else: df_this.to_csv(fout, sep="\t", index=False, header=bool(n_out==0))
    n_out+=1
  if tq is not None: tq.close()
  logging.info(f"n_out: {n_out}")
  return df

##############################################################################
def Study2DataFrame(study):
  # What to include? 
  protocol = study["protocolSection"]
  identification_module = protocol["identificationModule"]
  status_module = protocol["statusModule"]
  sponsor_collaborators_module = protocol["sponsorCollaboratorsModule"]
  description_module = protocol["descriptionModule"]
  conditions_module = protocol["conditionsModule"] if "conditionsModule" in protocol else None
  design_module = protocol["designModule"]
  arms_interventions_module = protocol["armsInterventionsModule"] if "armsInterventionsModule" in protocol else None
  outcomes_module = protocol["outcomesModule"] if "outcomesModule" in protocol else None
  eligibility_module = protocol["eligibilityModule"]
  oversight_module = protocol["oversightModule"] if "oversightModule" in protocol else None
  references_module = protocol["referencesModule"] if "referencesModule" in protocol else None
  
  derived = study["derivedSection"]
  condition_browse_module = derived["conditionBrowseModule"] if "conditionBrowseModule" in derived else None
  intervention_browse_module = derived["interventionBrowseModule"] if "interventionBrowseModule" in derived else None
  
  nct_id = identification_module["nctId"]
  brief_title = identification_module["briefTitle"]
  official_title = identification_module["officialTitle"] if "officialTitle" in identification_module else None
  
  organization = identification_module["organization"]
  org_name = organization["fullName"]
  org_class = organization["class"]
  
  has_results = study["hasResults"]
  
  df_this = pd.DataFrame({
          "nctId":[nct_id],
          "briefTitle":[brief_title],
          "officialTitle":[official_title],
          "orgName":[org_name],
          "orgClass":[org_class],
          "hasResults":[has_results]
          })
  return df_this

##############################################################################



#!/usr/bin/env python3
"""
Utility functions for HuBMAP REST API.
https://portal.hubmapconsortium.org/apis
"""
###
import sys,os,re,time,json,logging,requests,tqdm
import pandas as pd
#
API_HOST='entity.api.hubmapconsortium.org'
API_BASE_PATH=''
API_BASE_URL = f"https://{API_HOST}{API_BASE_PATH}"
#
#############################################################################
def ListEntityTypes(base_url=API_BASE_URL, fout=None):
  response = requests.get(f"{base_url}/entity-types")
  if response.status_code != 200:
    logging.error(f"status_code: {response.status_code}")
    return []
  results = response.json()
  if fout is not None:
    for result in results:
      fout.write(f"{result}\n")
  logging.info(f"Entities: {len(results)}")
  return results

#############################################################################
def GetEntity(ids, base_url=API_BASE_URL, fout=None):
  n_out=0; df=None; tq=None;
  for id_this in ids:
    response = requests.get(f"{base_url}/entities/{id_this}")
    if response.status_code != 200:
      logging.error(f"status_code: {response.status_code}")
      continue
    results = response.json()
    logging.debug(json.dumps(results, indent=2))
    if fout is not None:
      for result in results:
        fout.write(f"{result}\n")
    n_out += len(results)
  logging.info(f"n_out: {n_out}")
  return df

#############################################################################



#!/usr/bin/env python3
"""
Utility functions for Oncotree REST API.
See: https://oncotree.mskcc.org/
See: https://oncotree.mskcc.org/#/home
"""
import sys,os,re,requests,urllib,json,time,logging,tqdm
import pandas as pd

#
API_HOST="oncotree.mskcc.org"
API_BASE_PATH="/api"
BASE_URL="https://"+API_HOST+API_BASE_PATH
#
HEADERS={"Accept":"application/json"}
#
##############################################################################
def Info(base_url=BASE_URL, fout=None):
  response = requests.get(base_url+"/info", headers=HEADERS)
  result = response.json()
  logging.debug(json.dumps(result, sort_keys=True, indent=2))

##############################################################################
def ListVersions(base_url=BASE_URL, fout=None):
  response = requests.get(base_url+'/versions', headers=HEADERS)
  result = response.json()
  logging.debug(json.dumps(result, sort_keys=True, indent=2))
  versions = result
  df = pd.DataFrame.from_records(versions)
  if fout: df.to_csv(fout, sep="\t", index=False)
  logging.info(f"Versions: {len(versions)}")
  return df

##############################################################################
def ListMainTypes(base_url=BASE_URL, fout=None):
  response = requests.get(base_url+'/mainTypes', headers=HEADERS)
  result = response.json()
  logging.debug(json.dumps(result, sort_keys=True, indent=2))
  maintypes = result
  maintypes.sort()
  df = pd.DataFrame({"main_types": maintypes})
  if fout: df.to_csv(fout, sep="\t", index=False)
  logging.info(f"Main types: {len(maintypes)}")
  return df

##############################################################################
def ListTumorTypes(base_url=BASE_URL, fout=None):
  response = requests.get(base_url+'/tumorTypes', headers=HEADERS)
  result = response.json()
  logging.debug(json.dumps(result, sort_keys=True, indent=2))
  tumortypes = result
  df = pd.DataFrame.from_records(tumortypes)
  if fout: df.to_csv(fout, sep="\t", index=False)
  logging.info(f"Tumor types: {len(tumortypes)}")
  return df

##############################################################################
def SearchTumorTypes(qry, qtype, exact, levels, base_url=BASE_URL, fout=None):
  response = requests.get(base_url+f"/tumorTypes/search/{qtype}/{qry}?exactMatch={str(exact)}&levels={urllib.parse.quote(levels)}", headers=HEADERS)
  result = response.json()
  logging.debug(json.dumps(result, sort_keys=True, indent=2))
  tumortypes = result
  df = pd.DataFrame.from_records(tumortypes)
  if fout: df.to_csv(fout, sep="\t", index=False)
  logging.info(f"Tumor types: {len(tumortypes)}")
  return df

##############################################################################



#!/usr/bin/env python3
"""
Utility for HuBMAP REST API.
"""
###
import sys,os,re,json,argparse,time,logging
#
from .. import hubmap as hubmap_utils
#
##############################################################################
if __name__=='__main__':
  epilog="""
Example IDs:
HBM437.HTCQ.742 (donor)
HBM525.JNPV.685 (donor)
HBM292.WDQS.245 (sample)
HBM638.DVCH.366 (sample)
HBM543.RSRV.265 (dataset)
HBM287.WDKX.539 (dataset)
HBM925.SGXL.596 (collection)
HBM876.XNRH.336 (collection)
"""
  parser = argparse.ArgumentParser(description='HuBMAP REST API client', epilog=epilog)
  ops = ['list_entity_types', 'get_entity', 
    ]
  parser.add_argument("op", choices=ops, help='OPERATION')
  parser.add_argument("--ids", dest="ids", help="IDs, comma-separated")
  parser.add_argument("--i", dest="ifile", help="input file, HuBMAP entity IDs or UUIDs")
  parser.add_argument("--o", dest="ofile", help="output file (TSV)")
  parser.add_argument("--api_host", default=hubmap_utils.API_HOST)
  parser.add_argument("--api_base_path", default=hubmap_utils.API_BASE_PATH)
  parser.add_argument("-v", "--verbose", action="count", default=0)
  args = parser.parse_args()

  logging.basicConfig(format='%(levelname)s:%(message)s', level=(logging.DEBUG if args.verbose>1 else logging.INFO))

  base_url = f"https://{args.api_host}{args.api_base_path}"

  ids=[]
  if args.ifile:
    fin = open(args.ifile)
    while True:
      line = fin.readline()
      if not line: break
      ids.append(line.strip())
  elif args.ids:
    ids = re.split('[, ]+', args.ids.strip())

  fout = open(args.ofile, "w+") if args.ofile else sys.stdout

  t0=time.time()

  if args.op == "list_entity_types":
    hubmap_utils.ListEntityTypes(base_url, fout)

  elif args.op == "get_entity":
    hubmap_utils.GetEntity(ids, base_url, fout)

  else:
    parser.error(f"Invalid operation: {args.op}")

  logging.info(f"Elapsed time: {time.strftime('%Hh:%Mm:%Ss', time.gmtime(time.time()-t0))}")



#!/usr/bin/env python3
"""
Convert CSV to INSERTS for a specified table, with
control over column names, datatypes, and database systems (dbsystem).
"""
###
import sys,os,argparse,re,gzip,logging
import pandas as pd
#
CHUNKSIZE=100;
DBSYSTEMS=['postgres', 'mysql', 'oracle', 'derby']
DBSYSTEM='postgres'
KEYWORDS='procedure,function,column,table'
MAXCHAR=1024
CHARTYPES = 'CHAR,CHARACTER,VARCHAR,TEXT'
NUMTYPES = 'INT,BIGINT,FLOAT,NUM'
DEFTYPE='CHAR'
TIMETYPES = 'DATE,TIMESTAMP'
NULLWORDS = "NULL,UNSPECIFIED,MISSING,UNKNOWN,NA,NAN"
SCHEMA='public'
QUOTECHAR='"'
#
#############################################################################
def CsvCheck(fin, dbsystem, noheader, maxchar, delim, qc, keywords):
  colnames=None; n_in=0; n_err=0; i_chunk=0;
  with pd.read_csv(fin, chunksize=CHUNKSIZE, sep=delim) as reader:
    for df_this in reader:
      i_chunk+=1
      logging.debug(f"chunk: {i_chunk}; nrows: {df_this.shape[0]}")
      for i in range(df_this.shape[0]):
        row = df_this.iloc[i,] #Series
        if i_chunk==1 or colnames is None:
          if noheader:
            prefix = re.sub(r'\..*$', '', os.path.basename(fin.name))
            colnames = [f"{prefix}_{j}" for j in range(1, 1+len(row))]
          else:
            colnames = list(df_this.columns)
          colnames_clean = CleanNames(colnames, '', keywords)
          colnames_clean = DedupNames(colnames_clean)
          for j in range(len(colnames)):
            tag = colnames[j]
            tag_clean = colnames_clean[j]
            logging.debug(f"column tag {j+1}: {tag:>24}"+(f" -> {tag_clean}" if tag_clean!=tag else ""))
          nval = {colname:0 for colname in colnames}
          maxlen = {colname:0 for colname in colnames}
          dtypes = {colname:type(row.values[j]) for j,colname in enumerate(colnames)}
        for j in range(len(row)):
          val = list(row.values)[j]
          if j<=len(colnames):
            colname = colnames[j]
          else:
            logging.error(f"[{n_in+i+1}] row j_col>len(colnames) {j}>{len(colnames)}")
            n_err+=1
          if type(val) is str:
            try:
              #val = val.encode('utf-8', 'replace')
              if len(val.strip())>0: nval[colname]+=1
              val = EscapeString(val, dbsystem)
              maxlen[colname] = max(maxlen[colname], len(val))
              if len(val)>maxchar:
                logging.debug(f"WARNING: [{n_in+i+1}] len>MAX ({len(val)}>{maxchar})")
                val = val[:maxchar]
            except Exception as e:
              logging.error(f"[{n_in+i+1}] {str(e)}")
              val = f"CSV2SQL:ENCODING_ERROR"
              n_err+=1
          else:
              if val is not None: nval[colname]+=1
        n_in+=df_this.shape[0]
  for j,tag in enumerate(colnames):
    logging.info(f"""{j+1}. {tag:>24}: nval: {nval[tag]:6d}; type: {dtypes[tag]}; maxlen: {maxlen[tag]:6d}""")
  logging.info(f"n_in: {n_in}; n_err: {n_err}")

#############################################################################
def Csv2Create(fin, fout, dbsystem, dtypes, schema, tablename, colnames, coltypes, prefix, noheader, fixtags, delim, qc, keywords):
  n_in=0; n_err=0; i_chunk=0;
  with pd.read_csv(fin, chunksize=CHUNKSIZE, sep=delim, dtype=str) as reader:
    for df_this in reader:
      i_chunk+=1
      logging.debug(f"chunk: {i_chunk}; nrows: {df_this.shape[0]}")
      for i in range(df_this.shape[0]):
        row = df_this.iloc[i,] #Series
        if i_chunk==1:
          if colnames is None:
            if noheader:
              prefix = re.sub(r'\..*$', '', os.path.basename(fin.name))
              colnames = [f"{prefix}_{j}" for j in range(1, 1+len(row))]
            else:
              colnames = list(df_this.columns)
          else:
            if len(colnames) != len(df_this.columns):
              logging.error("#colnames)!=#df_this.columns ({len(colnames)}!={len(df_this.columns)})")
              return
          if fixtags:
            colnames_orig = colnames[:]
            colnames_clean = CleanNames(colnames, '', keywords)
            colnames = DedupNames(colnames_clean)
            for j in range(len(colnames)):
              logging.debug(f"column {j+1}: {colnames_orig[j]:>24}"+(f" -> {colnames[j]}" if colnames[j]!=colnames_orig[j] else ""))
          nval = {colname:0 for colname in colnames}
          maxlen = {colname:0 for colname in colnames}
        if coltypes is None:
          coltypes = [dtypes['deftype'] for i in range(len(colnames))]
        else:
          if len(coltypes) != len(df_this.columns):
            logging.error("#coltypes!=#df_this.columns ({len(coltypes)}!={len(df_this.columns)})")
            return
          for j in range(len(coltypes)):
            if coltypes[j] is None: coltypes[j] = 'CHAR'
  sql = f"CREATE TABLE {tablename} (\n\t" if dbsystem=='mysql' else f"CREATE TABLE {schema}.{tablename} (\n\t"
  sql+=(",\n\t".join([f"{colnames[j]} {dtypes[dbsystem][coltypes[j]]}" for j in range(len(colnames))]))
  sql+=("\n);")
  if dbsystem=='postgres':
    sql += f"\nCOMMENT ON TABLE {schema}.{tablename} IS 'Created by CSV2SQL.';"
  fout.write(sql+'\n')
  logging.info(f"{tablename}: output SQL CREATE written, columns: {len(colnames)}")

#############################################################################
def Csv2Insert(fin, fout, dbsystem, dtypes, schema, tablename, colnames, coltypes, prefix, noheader, nullwords, nullify, fixtags, maxchar, chartypes, numtypes, timetypes, delim, qc, keywords, skip, nmax):
  n_in=0; n_err=0; i_chunk=0; n_out=0;
  with pd.read_csv(fin, chunksize=CHUNKSIZE, sep=delim, dtype=str) as reader:
    for df_this in reader:
      i_chunk+=1
      logging.debug(f"chunk: {i_chunk}; nrows: {df_this.shape[0]}")
      for i in range(df_this.shape[0]):
        n_in+=1
        if i_chunk==1:
          if colnames is None:
            if noheader:
              prefix = re.sub(r'\..*$', '', os.path.basename(fin.name))
              colnames = [f"{prefix}_{j}" for j in range(1, 1+df_this.shape[1])]
            else:
              colnames = list(df_this.columns)
            logging.debug(f"columns: {colnames}")
          else:
            if len(colnames) != len(df_this.columns):
              logging.error("#colnames)!=#df_this.columns ({len(colnames)}!={len(df_this.columns)})")
              return
          if coltypes is None:
            coltypes = [dtypes['deftype'] for i in range(len(colnames))]
          else:
            if len(coltypes)!=len(colnames):
              logging.error(f"#coltypes!=#colnames ({len(coltypes)}!={len(colnames)})")
              return
          if fixtags:
            colnames_orig = colnames[:]
            colnames_clean = CleanNames(colnames, '', keywords)
            colnames = DedupNames(colnames_clean)
            for j in range(len(colnames)):
              logging.debug(f"column {j+1}: {colnames_orig[j]:>24}"+(f" -> {colnames[j]}" if colnames[j]!=colnames_orig[j] else ""))
            logging.debug(f"columns: {colnames}")
        if n_in<=skip: continue
        line = (f"INSERT INTO {tablename} ({','.join(colnames)}) VALUES (") if dbsystem=='mysql' else (f"INSERT INTO {schema}.{tablename} ({','.join(colnames)}) VALUES (")
        df_this.columns = colnames
        row = df_this.iloc[i,] #Series
        logging.debug(f"row.keys(): {row.keys()}")
        for j,colname in enumerate(colnames):
          val = str(row[colname])
          if coltypes[j].upper() in chartypes:
            val = EscapeString(str(val), dbsystem)
            if len(val)>maxchar:
              val = val[:maxchar]
              logging.info(f"WARNING: [row={n_in}] string truncated to {maxchar} chars: '{val}'")
            val = "NULL" if (nullify and val.upper() in nullwords) else f"'{val}'"
          elif coltypes[j].upper() in numtypes:
            val = "NULL" if (val.upper() in nullwords or val=='') else f"{val}"
          elif coltypes[j].upper() in timetypes:
            val = (f"to_timestamp('{val}')")
          else:
            logging.error(f"No type specified or implied: (col={j+1})")
            continue
          line+=(f"{',' if j>0 else ''}{val}")
        line +=(') ;')
        fout.write(line+'\n')
        n_out+=1
        if n_in==nmax: break
  logging.info(f"{tablename}: input CSV lines: {n_in}; output SQL inserts: {n_out}")

#############################################################################
def CleanName(name, keywords):
  '''Clean table or col name for use without escaping.
1. Downcase.
2. Replace spaces and colons with underscores.
3. Remove punctuation and special chars.
4. Prepend leading numeral.
5. Truncate to 50 chars.
6. Fix if in keyword list, e.g. "procedure".
'''
  name = re.sub(r'[\s:-]+','_',name.lower())
  name = re.sub(r'[^\w]','',name)
  name = re.sub(r'^([\d])',r'col_\1',name)
  name = name[:50]
  if name in keywords:
    name+='_name'
  return name

#############################################################################
def CleanNames(colnames, prefix, keywords):
  for j in range(len(colnames)):
    colnames[j] = CleanName(prefix+colnames[j] if prefix else colnames[j], keywords)
  return colnames

#############################################################################
def DedupNames(colnames):
  unames = set()
  for j in range(len(colnames)):
    if colnames[j] in unames:
      colname_orig = colnames[j]
      k=1
      while colnames[j] in unames:
        k+=1
        colnames[j] = f"{colname_orig}_{k}"
    unames.add(colnames[j])
  return colnames

#############################################################################
def EscapeString(val, dbsystem):
  val=re.sub(r"'", '', val)
  if dbsystem=='postgres':
    val=re.sub(r'\\', r"'||E'\\\\'||'", val)
  elif dbsystem=='mysql':
    val=re.sub(r'\\', r'\\\\', val)
  return val

#############################################################################
if __name__=='__main__':

  parser = argparse.ArgumentParser(description="CSV to SQL CREATEs or INSERTs", epilog='')
  ops = ["insert", "create", "check"]
  parser.add_argument("op", choices=ops, help='OPERATION')
  parser.add_argument("--i", dest="ifile", help="input file (CSV|TSV) [stdin]")
  parser.add_argument("--o", dest="ofile", help="output SQL INSERTs [stdout]")
  parser.add_argument("--schema", default=SCHEMA, help="(Postgres schema, or MySql db)")
  parser.add_argument("--tablename", help="table [convert filename]")
  parser.add_argument("--prefix_tablename", help="prefix + [convert filename]")
  parser.add_argument("--dbsystem", default=DBSYSTEM, help="")
  parser.add_argument("--colnames", help="comma-separated [default: CSV tags]")
  parser.add_argument("--noheader", action="store_true", help="auto-name columns")
  parser.add_argument("--prefix_colnames", help="prefix CSV tags")
  parser.add_argument("--coltypes", help="comma-separated [default: all CHAR]")
  parser.add_argument("--nullify", action="store_true", help="CSV missing CHAR value converts to NULL")
  parser.add_argument("--nullwords", default=NULLWORDS, help="words synonymous with NULL (comma-separated list)")
  parser.add_argument("--keywords", default=KEYWORDS, help="keywords, disallowed tablenames (comma-separated list)")
  parser.add_argument("--char_types", default=CHARTYPES, help="character types (comma-separated list)")
  parser.add_argument("--num_types", default=NUMTYPES, help="numeric types (comma-separated list)")
  parser.add_argument("--time_types", default=TIMETYPES, help="time types (comma-separated list)")
  parser.add_argument("--maxchar", type=int, default=MAXCHAR, help="max string length")
  parser.add_argument("--fixtags", action="store_true", help="tags to colnames (downcase/nopunct/nospace)")
  parser.add_argument("--tsv", action="store_true", help="input file TSV")
  parser.add_argument("--igz", action="store_true", help="input file GZ")
  parser.add_argument("--delim", default=",", help="use if not comma or tab")
  parser.add_argument("--default_type", default=DEFTYPE, help="")
  parser.add_argument("--quotechar", default=QUOTECHAR, help="")
  parser.add_argument("--skip", type=int, default=0, help="skip N records (--insert only)")
  parser.add_argument("--nmax", type=int, default=0, help="max N records (--insert only)")
  parser.add_argument("-v", "--verbose", default=0, action="count")
  args = parser.parse_args()

  logging.basicConfig(format='%(levelname)s:%(message)s', level=(logging.DEBUG if args.verbose>1 else logging.INFO))

  DELIM = '\t' if args.tsv else args.delim

  NULLWORDS = re.split(r'[\s,]', args.nullwords)
  KEYWORDS = re.split(r'[\s,]', args.keywords)
  CHARTYPES = re.split(r'[\s,]', args.char_types)
  NUMTYPES = re.split(r'[\s,]', args.num_types)
  TIMETYPES = re.split(r'[\s,]', args.time_types)

  fin = (gzip.open(args.ifile) if args.igz else open(args.ifile)) if args.ifile else sys.stdin
  fout = open(args.ofile,'w') if args.ofile else sys.stdout

  if args.tablename:
    TABLENAME = args.tablename
  else:
    if not args.ifile: parser.error('--tablename or --i required.')
    TABLENAME = CleanName(args.prefix_tablename+re.sub(r'\..*$','', os.path.basename(args.ifile)), KEYWORDS)
    logging.debug(f"""tablename = '{TABLENAME}'""")

  DTYPES = {
	"postgres": {
		"CHAR":f"VARCHAR({args.maxchar})",
		"CHARACTER":f"VARCHAR({args.maxchar})",
		"VARCHAR":f"VARCHAR({args.maxchar})",
		"INT":"INTEGER",
		"BIGINT":"BIGINT",
		"FLOAT":"FLOAT",
		"NUM":"FLOAT",
		"BOOL":"BOOLEAN"
		},
	"mysql": {
		"CHAR":f"TEXT({args.maxchar})",
		"CHARACTER":f"TEXT({args.maxchar})",
		"VARCHAR":f"TEXT({args.maxchar})",
		"INT":"INTEGER",
		"BIGINT":"BIGINT",
		"FLOAT":"FLOAT",
		"NUM":"FLOAT",
		"BOOL":"BOOLEAN"
		}
	}
  DTYPES["deftype"] = args.default_type

  if args.op=="create":
    Csv2Create(fin, fout, args.dbsystem,
	DTYPES,
	args.schema,
	TABLENAME,
	re.split(r'[,\s]+', args.colnames) if args.colnames else None,
	re.split(r'[,\s]+', args.coltypes) if args.coltypes else None,
	args.prefix_colnames, args.noheader, args.fixtags, DELIM, args.quotechar,
	KEYWORDS)

  elif args.op=="insert":
    Csv2Insert(fin, fout, args.dbsystem,
	DTYPES,
	args.schema,
	TABLENAME,
	re.split(r'[,\s]+', args.colnames) if args.colnames else None,
	re.split(r'[,\s]+', args.coltypes) if args.coltypes else None,
	args.prefix_colnames, args.noheader,
	NULLWORDS, args.nullify, args.fixtags, args.maxchar,
	CHARTYPES, NUMTYPES, TIMETYPES,
	DELIM, args.quotechar, KEYWORDS, args.skip, args.nmax)

  elif args.op=="check":
    CsvCheck(fin, args.dbsystem, args.noheader, args.maxchar, DELIM, args.quotechar, KEYWORDS)

  else:
    parser.error(f"Invalid operation: {args.op}")




#!/usr/bin/env python3
##############################################################################
### OpenPHACTS REST API Client
### https://dev.openphacts.org/docs
### https://dev.openphacts.org/docs/1.5
### https://dev.openphacts.org/docs/2.0
##############################################################################
### From https://dev.openphacts.org/docs/develop
#  ?oidd_assay_uri a ?bao_type ;
#     skos:exactMatch ?pubchem_assay ;
#     void:inDataset <http://rdf.ncats.nih.gov/opddr> .
###
#  /assay?uri=http%3A%2F%2Fopeninnovation.lilly.com%2Fbioassay%2329"
###
##############################################################################
import sys,os,re,time,argparse,logging
#
from .. import openphacts
#
from ..util import yaml as util_yaml
#
#
##############################################################################
if __name__=='__main__':
  epilog='''\
Examples:
http://openinnovation.lilly.com/bioassay#29 (assay)
http://www.conceptwiki.org/concept/38932552-111f-4a4e-a46a-4ed1d7bdf9d5 (compound)
http://www.conceptwiki.org/concept/00059958-a045-4581-9dc5-e5a08bb0c291 (target)
'''
  parser = argparse.ArgumentParser(description='ClinicalTrials.gov API client', epilog=epilog)
  ops = [ "counts", "get_assay", "get_cpd", "get_target", "list_pathways", "list_sources", "list_tgt_types", "list_targets", ]
  parser.add_argument("op", choices=ops, help='OPERATION')
  parser.add_argument("--qry_uri", help="query URI")
  parser.add_argument("--qry_txt", help="query text")
  parser.add_argument("--o", dest="ofile", help="Output (TSV)")
  parser.add_argument("--api_host", default=openphacts.API_HOST)
  parser.add_argument("--api_base_path", default=openphacts.API_BASE_PATH)
  parser.add_argument("--api_user_id")
  parser.add_argument("--api_user_key")
  parser.add_argument("--param_file", default=os.environ['HOME']+"/.openphacts.yaml")
  parser.add_argument("-v", "--verbose", default=0, action="count")
  args = parser.parse_args()

  logging.basicConfig(format='%(levelname)s:%(message)s', level=(logging.DEBUG if args.verbose>1 else logging.INFO))

  params = util_yaml.ReadParamFile(args.param_file) if os.path.isfile(args.param_file) else {}
  api_user_id = args.api_user_id if args.api_user_id else params['ID'] if 'ID' in params else None
  api_user_key = args.api_user_key if args.api_user_key else params['KEY'] if 'KEY' in params else None

  base_url = 'https://'+args.api_host+args.api_base_path

  fout = open(ofile, "w+") if args.ofile else sys.stdout

  if args.op == "list_pathways":
    openphacts.ListPathways(api_user_id, api_user_key, base_url, fout)

  elif args.op == "list_sources":
    openphacts.ListSources(api_user_id, api_user_key, base_url, fout)

  elif args.op == "list_tgt_types":
    openphacts.ListTargetTypes(api_user_id, api_user_key, base_url, fout)

  elif args.op == "list_targets":
    if not qry_uri: parser.error('Query URI required.')
    openphacts.ListTargets(qry_uri, api_user_id, api_user_key, base_url, fout)

  elif args.op == "counts":
    openphacts.ShowCounts(api_user_id, api_user_key, base_url)

  elif args.op == "get_assay":
    if not qry_uri: parser.error('Query URI required.')
    openphacts.GetInstance('assay', qry_uri, api_user_id, api_user_key, base_url, fout)

  elif args.op == "get_cpd":
    if not qry_uri: parser.error('Query URI required.')
    openphacts.GetInstance('compound', qry_uri, api_user_id, api_user_key, base_url, fout)

  elif args.op == "get_target":
    if not qry_uri: parser.error('Query URI required.')
    openphacts.GetInstance('target', qry_uri, api_user_id, api_user_key, base_url, fout)

  else:
    parser.error(f"Operation invalid: {args.op}")



#!/usr/bin/env python3
"""
ClinicalTrials.gov
https://clinicaltrials.gov/api/gui/ref/api_urls
API v2.0: https://clinicaltrials.gov/data-api/api
https://clinicaltrials.gov/data-api/about-api/csv-download
"""
### 
import sys,os,re,json,argparse,time,logging

from .. import clinicaltrials
#
#############################################################################
if __name__=='__main__':
  epilog='''\
See:
https://clinicaltrials.gov/data-api/about-api, 
https://clinicaltrials.gov/data-api/api, 
https://clinicaltrials.gov/find-studies/constructing-complex-search-queries
'''
  parser = argparse.ArgumentParser(description='ClinicalTrials.gov API client', epilog=epilog)
  ops = [ 'version',
	'list_study_fields', 'list_search_areas', 'list_enums',
	'search_studies', 'get_studies', ]
  parser.add_argument("op", choices=ops, help='OPERATION')
  parser.add_argument("--i", dest="ifile", help="Input NCT_IDs")
  parser.add_argument("--o", dest="ofile", help="Output (TSV)")
  parser.add_argument("--ids", help="Input NCT_IDs (comma-separated)")
  parser.add_argument("--query_cond", help="Search query condition")
  parser.add_argument("--query_term", help="Search query term")
  parser.add_argument("--api_host", default=clinicaltrials.API_HOST)
  parser.add_argument("--api_base_path", default=clinicaltrials.API_BASE_PATH)
  parser.add_argument("-v", "--verbose", default=0, action="count")
  args = parser.parse_args()

  logging.basicConfig(format='%(levelname)s:%(message)s', level=(logging.DEBUG if args.verbose>1 else logging.INFO))

  api_base_url = 'https://'+args.api_host+args.api_base_path

  fout = open(args.ofile, "w+") if args.ofile else sys.stdout

  t0=time.time()

  ids=[]
  if args.ifile:
    fin = open(args.ifile)
    while True:
      line = fin.readline()
      if not line: break
      ids.append(line.rstrip())
    fin.close()
    logging.info(f"Input IDs: {len(ids)}")
  elif args.ids:
    ids = re.split(r'[,\s]+', args.ids)

  if args.op == "version":
    clinicaltrials.Version(api_base_url, fout)

  elif args.op == "list_study_fields":
    clinicaltrials.ListStudyFields(api_base_url, fout)

  elif args.op == "list_search_areas":
    clinicaltrials.ListSearchAreas(api_base_url, fout)

  elif args.op == "list_enums":
    clinicaltrials.ListEnums(api_base_url, fout)

  elif args.op == "search_studies":
    clinicaltrials.SearchStudies(args.query_cond, args.query_term, api_base_url, fout)

  elif args.op == "get_studies":
    clinicaltrials.GetStudies(ids, api_base_url, fout)

  else:
    parser.error(f"Invalid operation: {args.op}")

  logging.info(f"Elapsed time: {time.strftime('%Hh:%Mm:%Ss',time.gmtime(time.time()-t0))}")



#!/usr/bin/env python3
"""
Utility functions for HUGO REST API.
See: https://www.genenames.org/help/rest-web-service-help
"""
import sys,os,re,requests,json,time,logging,tqdm
import pandas as pd

#
API_HOST="rest.genenames.org"
API_BASE_PATH=""
BASE_URL="http://"+API_HOST+API_BASE_PATH
#
HEADERS={"Accept":"application/json"}
#
##############################################################################
def Info(base_url=BASE_URL, fout=None):
  response = requests.get(base_url+"/info", headers=HEADERS)
  result = response.json()
  logging.debug(json.dumps(result, sort_keys=True, indent=2))
  if "searchableFields" in result:
    for f in result["searchableFields"]:
      print(f"{f}")
  if "storedFields" in result:
    for f in result["storedFields"]:
      print(f"{f}")

##############################################################################
def ListSearchableFields(base_url=BASE_URL, fout=None):
  response = requests.get(base_url+'/info', headers=HEADERS)
  result = response.json()
  fields = result['searchableFields'] if 'searchableFields' in result else None
  fields.sort()
  df = pd.DataFrame({'fields':fields})
  if fout: df.to_csv(fout, sep="\t", index=False)
  return df

##############################################################################
def ListStoredFields(base_url=BASE_URL, fout=None):
  response = requests.get(base_url+'/info', headers=HEADERS)
  result = response.json()
  fields = result['storedFields'] if 'storedFields' in result else None
  fields.sort()
  df = pd.DataFrame({'fields':fields})
  if fout: df.to_csv(fout, sep="\t", index=False)
  return df

##############################################################################
def GetGenes(qrys, ftypes, skip=0, base_url=BASE_URL, fout=None):
  '''The API can return multiple hits for each query, though should not 
for exact SYMBOL fetch.  One case is for RPS15P5, status "Entry Withdrawn".'''
  n_in=0; n_found=0; n_notfound=0; n_ambig=0; n_out=0;
  tags=None; df=None; tq=None;
  logging.debug(f"ftypes[{len(ftypes)}]: {str(ftypes)}")
  for qry in qrys:
    if not tq: tq = tqdm.tqdm(total=len(qrys)-skip)
    tq.update()
    n_in+=1
    if n_in<skip: continue
    logging.debug(f"{n_in}. query: {qry}")
    numFound=0; genes=[];
    ftype_hit=None
    for ftype in ftypes:
      response = requests.get(base_url+f"/fetch/{ftype.lower()}/{qry}", headers=HEADERS)
      result = response.json()
      logging.debug(json.dumps(result, indent=2))
      if result is None: continue
      ftype_hit=ftype
      numFound = result['response']['numFound'] if 'numFound' in result['response'] else 0
      genes = result['response']['docs']
      if numFound>0:
        logging.debug(f"Not found: {ftype_hit} = {qry}")
        break
    if numFound==0:
      n_notfound+=1
      logging.debug(f"Not found: {ftype_hit} = {qry}")
      continue
    elif numFound>0:
      n_found+=1
    if len(genes)>1:
      logging.warning(f"Multiple ({len(genes)}) hits, {ftype_hit} = {qry} (Duplicate may be status=withdrawn.)")
      n_ambig+=1
    for gene in genes:
      gene_data = {'query':[qry], 'field':[ftype_hit]}
      for tag in gene.keys():
        if tag in ("prev_symbol", "prev_name", "alias_symbol"):
          gene_data[tag] = [";".join(gene[tag])]
        elif type(gene[tag]) in (list, dict):
          continue
        elif tags is not None and tag not in tags:
          continue
        else:
          gene_data[tag] = [gene[tag]]
      if not tags:
        tags = gene_data.keys()
        logging.debug(f"tags: {tags}")
      for tag in (set(tags) - set(gene_data.keys())):
        gene_data[tag] = [None] #add missing fields
      df_this = pd.DataFrame(gene_data)
      df_this = df_this[tags] #reorder cols
      if fout: df_this.to_csv(fout, sep="\t", index=False, header=bool(n_out==0))
      else: df = pd.concat([df, df_this])
      n_out+=1
  if tq is not None: tq.close()
  logging.info(f"queries: {n_in-skip}; n_found: {n_found}; n_notfound: {n_notfound}; n_ambig: {n_ambig}; n_out: {n_out}")
  return df

##############################################################################
def SearchGenes(qrys, ftypes, base_url=BASE_URL, fout=None):
  n_in=0; n_found=0; tags=[]; df=None;
  for qry in qrys:
    n_in+=1
    logging.debug(f"{n_in}. query: {qry}")
    found_this=False
    for ftype in ftypes:
      url = (base_url+'/search{}/*{}*'.format((('/'+ftype.lower()) if ftype else ''), qry))
      response = requests.get(url, headers=HEADERS)
      result = response.json()
      numFound = result['response']['numFound'] if 'numFound' in result['response'] else 0
      if numFound==0: continue
      found_this=True
      docs = result['response']['docs'] if 'docs' in result['response'] else []
      for doc in docs:
        logging.debug(json.dumps(doc, sort_keys=True, indent=2)+'\n')
        if not tags: tags = list(doc.keys())
        data_this = {'query':qry, 'field':ftype}
        data_this.update({tags[j]:[doc[tags[j]]] for j in range(len(tags))})
        df = pd.concat([df, pd.DataFrame(data_this)])
    if found_this: n_found+=1
  if fout: df.to_csv(fout, sep="\t", index=False)
  logging.info(f"queries: {n_in}; found: {n_found}")
  return df




#!/usr/bin/env python3
"""
Utility functions for GTEx REST API.
https://www.gtexportal.org/home/api-docs

curl -X GET "https://gtexportal.org/rest/v1/metadata/dataset?format=json&sortBy=datasetId" -H "accept: application/json"
"""
###
import sys,os,re,json,collections,time,urllib.parse,logging,tqdm,tqdm.auto
import pandas as pd
import requests
#
NCHUNK=100
#
API_HOST="gtexportal.org"
API_BASE_PATH="/rest/v1"
BASE_URL="https://"+API_HOST+API_BASE_PATH
#
##############################################################################
def ListDatasets(base_url=BASE_URL, fout=None):
  n_out=0; tags=None; df=None;
  response = requests.get(f"{base_url}/metadata/dataset?format=json&sortBy=datasetId", headers={"Content-Type":"application/json"})
  result = response.json()
  logging.debug(json.dumps(result, indent=2))
  datasets = result["dataset"]
  for dataset in datasets:
    logging.info(f"datasetId: {dataset['datasetId']}")
    if not tags: tags = [tag for tag in dataset.keys() if type(dataset[tag]) not in (list, dict, collections.OrderedDict)]
    df_this = pd.DataFrame({tags[j]:[dataset[tags[j]]] for j in range(len(tags))})
    if fout is None: df = pd.concat([df, df_this])
    else: df_this.to_csv(fout, sep="\t", index=False, header=bool(n_out==0))
    n_out += df_this.shape[0]
  logging.info(f"n_out: {n_out}")
  return df

##############################################################################
def ListSubjects(datasetId, base_url=BASE_URL, fout=None):
  n_out=0; tags=None; df=None; page=0;
  while True:
    url_this = f"{base_url}/dataset/subject?format=json&datasetId={datasetId}&pageSize={NCHUNK}&page={page}"
    response = requests.get(url_this, headers={"Content-Type":"application/json"})
    result = response.json()
    subjects = result["subject"]
    for subject in subjects:
      if not tags: tags = [tag for tag in subject.keys() if type(subject[tag]) not in (list, dict, collections.OrderedDict)]
      df_this = pd.DataFrame({tags[j]:[subject[tags[j]]] for j in range(len(tags))})
      if fout is None: df = pd.concat([df, df_this])
      else: df_this.to_csv(fout, sep="\t", index=False, header=bool(n_out==0))
      n_out += df_this.shape[0]
    if "numPages" in result and result["numPages"]-1>page:
      page+=1
    else:
      break
  logging.info(f"n_out: {n_out}")
  return df
 
##############################################################################
def ListSamples(datasetId, subjectId, base_url=BASE_URL, fout=None):
  n_out=0; tags=None; df=None; page=0;
  while True:
    url_this = f"{base_url}/dataset/sample?format=json&datasetId={datasetId}&subjectId={subjectId}&pageSize={NCHUNK}&page={page}"
    response = requests.get(url_this, headers={"Content-Type":"application/json"})
    result = response.json()
    samples = result["sample"]
    for sample in samples:
      if not tags: tags = [tag for tag in sample.keys() if type(sample[tag]) not in (list, dict, collections.OrderedDict)]
      df_this = pd.DataFrame({tags[j]:[sample[tags[j]]] for j in range(len(tags))})
      if fout is None: df = pd.concat([df, df_this])
      else: df_this.to_csv(fout, sep="\t", index=False, header=bool(n_out==0))
      n_out += df_this.shape[0]
    if "numPages" in result and result["numPages"]-1>page:
      page+=1
    else:
      break
  logging.info(f"n_out: {n_out}")
  return df
 
##############################################################################
# https://gtexportal.org/rest/v1/expression/geneExpression?datasetId=gtex_v8&gencodeId=ENSG00000188906.14&format=json
def GetGeneExpression(ids, datasetId, skip, base_url=BASE_URL, fout=None):
  n_out=0; tags=None; df=None;
  for i in tqdm.auto.trange(len(ids), desc="IDs"):
    id_this = ids[i]
    response = requests.get(f"{base_url}/expression/geneExpression?datasetId=gtex_v8&gencodeId={id_this}&format=json", headers={"Content-Type":"application/json"})
    result = response.json()
    #logging.debug(json.dumps(result, indent=2))
    gexs = result["geneExpression"]
    for gex in gexs:
      if not tags: tags = [tag for tag in gex.keys() if type(gex[tag]) not in (list, dict, collections.OrderedDict)]
      df_this = pd.DataFrame({tags[j]:[gex[tags[j]]] for j in range(len(tags))})
      if fout is None: df = pd.concat([df, df_this])
      else: df_this.to_csv(fout, sep="\t", index=False, header=bool(n_out==0))
      n_out += df_this.shape[0]
  logging.info(f"n_out: {n_out}")
  return df

##############################################################################



#!/usr/bin/env python3
"""
OWL utility functions.
https://owlready2.readthedocs.io/
"""
import sys,os,re,gzip,argparse,logging

import owlready2

#############################################################################
def LoadOwlFile(ifile):
  try:
    onto = owlready2.get_ontology(f"file://{ifile}").load()
    logging.info(f"OWL ontology from {ifile} contains {len(list(onto.classes()))} classes.")
  except Exception as e:
    logging.error(e)
    return None
  return onto

#############################################################################
def DescribeOwl(ifile):
  onto = LoadOwlFile(ifile)

#############################################################################
def ValidateOwl(ifile):
  onto = LoadOwlFile(ifile)
  if onto is not None:
    logging.info(f"OWL file VALIDATED: {ifile}")
    return True
  else:
    logging.info(f"OWL file NOT VALIDATED: {ifile}")
    return False

#############################################################################



#!/usr/bin/env python3
"""
Utility for STRING Db REST API.

STRING = Search Tool for the Retrieval of Interacting Genes/Proteins

See: http://string-db.org/help/api/

http://[database]/[access]/[format]/[request]?[parameter]=[value]

database:
string-db.org
string.embl.de
stitch.embl.de
"""
###
import sys,os,re,json,time,logging
import pandas as pd
import requests,urllib,urllib.request,urllib.parse
#
API_HOST='string-db.org'
API_BASE_PATH='/api'
BASE_URL='https://'+API_HOST+API_BASE_PATH
#
NETWORK_FLAVORS=['evidence', 'confidence', 'actions']
IMG_FMTS=['image', 'highres_image', 'svg']
#
##############################################################################
def GetIds(ids, base_url=BASE_URL, fout=None):
  tags=[]; df=pd.DataFrame();
  for id_this in ids:
    url_this = f"{base_url}/json/get_string_ids?identifier={id_this}"
    response = requests.get(url_this)
    results = response.json()
    for result in results:
      logging.debug(result)
      if not tags: tags = list(result.keys())
      df = pd.concat([df, pd.DataFrame({tags[j]:[result[tags[j]]] for j in range(len(tags))})])
  if fout: df.to_csv(fout, sep="\t", index=False)
  logging.info(f"queries: {len(ids)}; results: {df.shape[0]}")
  return df

##############################################################################
def GetInteractionPartners(ids, species, limit, minscore, base_url=BASE_URL, fout=None):
  """ALL interacting proteins, up to limit (ordered by confidence)"""
  tags=[]; df=pd.DataFrame();
  for id_this in ids:
    url_this = f"{base_url}/json/interaction_partners?identifier={id_this}"
    if species: url_this+=(f'&species={species}')
    if limit: url_this+=(f'&limit={limit}')
    if minscore: url_this+=(f'&required_score={minscore}')
    response = requests.get(url_this)
    results = response.json()
    for result in results:
      logging.debug(result)
      if not tags: tags = list(result.keys())
      df = pd.concat([df, pd.DataFrame({tags[j]:[result[tags[j]]] for j in range(len(tags))})])
  if fout: df.to_csv(fout, sep="\t", index=False)
  logging.info(f"queries: {len(ids)}; interaction partners: {df.shape[0]}")
  return df

##############################################################################
def GetEnrichment(ids, species, minscore, base_url=BASE_URL, fout=None):
  tags=[]; df=pd.DataFrame();
  ids_param = urllib.parse.quote('\n'.join(ids))
  url = f"{base_url}/json/enrichment?identifiers={ids_param}"
  if species: url+=(f'&species={species}')
  if minscore: url+=(f'&required_score={minscore}')
  response = requests.get(url)
  results = response.json()
  for result in results:
    logging.debug(result)
    if not tags: tags = list(result.keys())
    df = pd.concat([df, pd.DataFrame({tags[j]:[result[tags[j]]] for j in range(len(tags))})])
  if fout: df.to_csv(fout, sep="\t", index=False)
  logging.info(f"queries: {len(ids)}; enrichment results: {df.shape[0]}")
  return df

##############################################################################
def GetPPIEnrichment(ids, species, minscore, base_url=BASE_URL, fout=None):
  tags=[]; df=pd.DataFrame();
  ids_param = urllib.parse.quote('\n'.join(ids))
  url = f"{base_url}/json/ppi_enrichment?identifiers={ids_param}"
  if species: url+=(f'&species={species}')
  if minscore: url+=(f'&required_score={minscore}')
  response = requests.get(url)
  results = response.json()
  for result in results:
    logging.debug(result)
    if not tags: tags = list(result.keys())
    df = pd.concat([df, pd.DataFrame({tags[j]:[result[tags[j]]] for j in range(len(tags))})])
  if fout: df.to_csv(fout, sep="\t", index=False)
  logging.info(f"queries: {len(ids)}; enrichment results: {df.shape[0]}")
  return df

##############################################################################
def GetNetwork(nid, species, minscore, netflavor, base_url=BASE_URL, fout=None):
  tags=[]; df=pd.DataFrame();
  url = base_url+f'/json/network?identifier={nid}'
  if species: url+=(f'&species={species}')
  if minscore: url+=(f'&required_score={minscore}')
  if netflavor: url+=(f'&network_flavor={netflavor}')
  response = requests.get(url)
  edges = response.json()
  logging.debug(json.dumps(edges, indent=2))
  for edge in edges:
    logging.debug(edge)
    if not tags: tags = list(edge.keys())
    df = pd.concat([df, pd.DataFrame({tags[j]:[edge[tags[j]]] for j in range(len(tags))})])
  if fout: df.to_csv(fout, sep="\t", index=False)
  logging.info(f"edges: {df.shape[0]}")
  return df

##############################################################################
def GetNetworkImage(nid, species, minscore, netflavor, imgfmt, base_url=BASE_URL, fout=None):
  """image, highres_image, or svg"""
  url = base_url+f'/{imgfmt}/network?identifier={nid}'
  if species: url+=(f'&species={species}')
  if minscore: url+=(f'&required_score={minscore}')
  if netflavor: url+=(f'&network_flavor={netflavor}')
  response = requests.get(url)
  img = response.content
  if fout is not None:
    fout.write(img)
    fout.close()
  return img

##############################################################################



#!/usr/bin/env python3
"""
https://ncats.nih.gov/expertise/preclinical/gsrs
https://gsrs.ncats.nih.gov/
https://gsrs.ncats.nih.gov/#/api
"""
### 
import sys,os,re,json,argparse,time,logging

from ... import ncats
#
#############################################################################
if __name__=='__main__':
  epilog='''\
Example search queries:
IBUPRO
ASPIRIN
OXYTOCIN
OXYTO*
ASPIRIN AND ESTER
COCN
C=1CC=CC=C1C(=O)O
'''
  parser = argparse.ArgumentParser(description='NCATS Global Substance Registration System (GSRS) client', epilog=epilog)
  ops = [
	'list_vocabularies',
	'list_substances',
	'search',
	'get_substance',
	'get_substance_names',
	]
  parser.add_argument("op", choices=ops, help='OPERATION')
  parser.add_argument("--i", dest="ifile", help="Input IDs")
  parser.add_argument("--o", dest="ofile", help="Output (TSV)")
  parser.add_argument("--ids", help="Input IDs (comma-separated)")
  parser.add_argument("--query", help="Search query.")
  parser.add_argument("--api_host", default=ncats.gsrs.API_HOST)
  parser.add_argument("--api_base_path", default=ncats.gsrs.API_BASE_PATH)
  parser.add_argument("-v", "--verbose", default=0, action="count")
  args = parser.parse_args()

  logging.basicConfig(format='%(levelname)s:%(message)s', level=(logging.DEBUG if args.verbose>1 else logging.INFO))

  api_base_url = 'https://'+args.api_host+args.api_base_path

  fout = open(args.ofile, "w+") if args.ofile else sys.stdout

  t0=time.time()

  ids=[]
  if args.ifile:
    fin = open(args.ifile)
    while True:
      line = fin.readline()
      if not line: break
      ids.append(line.rstrip())
    fin.close()
    logging.info(f"Input IDs: {len(ids)}")
  elif args.ids:
    ids = re.split(r'[,\s]+', args.ids)

  if args.op == "list_vocabularies":
    ncats.gsrs.Utils.ListVocabularies(api_base_url, fout)

  elif args.op == "list_substances":
    ncats.gsrs.Utils.ListSubstances(api_base_url, fout)

  elif args.op == "search":
    ncats.gsrs.Utils.Search(args.query, api_base_url, fout)

  elif args.op == "get_substance":
    ncats.gsrs.Utils.GetSubstance(ids, api_base_url, fout)

  elif args.op == "get_substance_names":
    ncats.gsrs.Utils.GetSubstanceNames(ids, api_base_url, fout)

  else:
    parser.error(f"Invalid operation: {args.op}")

  logging.info(f"Elapsed time: {time.strftime('%Hh:%Mm:%Ss',time.gmtime(time.time()-t0))}")



#!/usr/bin/env python3
"""
Client for Bioregistry REST API.

See: https://bioregistry.io/apidocs/
"""
###
import sys,os,re,json,argparse,time,logging
#
from .. import bioregistry
#
##############################################################################
if __name__=='__main__':
  parser = argparse.ArgumentParser(description='Bioregistry REST API client', epilog='')
  ops = [
	'list_collections', 
	'list_contexts', 
	'list_registry', 
	'list_metaregistry', 
	'list_contributors', 
	'get_reference', 
	]
  parser.add_argument("op", choices=ops, help='operation')
  parser.add_argument("--i", dest="ifile", help="input query IDs")
  parser.add_argument("--ids", help="input query IDs (comma-separated)")
  parser.add_argument("--o", dest="ofile", help="output (TSV)")
  parser.add_argument("--etype", default="", help="evidence codes (|-separated)")
  parser.add_argument("--prefix", help="CURIE prefix")
  parser.add_argument("--nchunk", type=int)
  parser.add_argument("--nmax", type=int, default=None)
  parser.add_argument("--skip", type=int, default=0)
  parser.add_argument("--api_host", default=bioregistry.API_HOST)
  parser.add_argument("--api_base_path", default=bioregistry.API_BASE_PATH)
  parser.add_argument("-v", "--verbose", default=0, action="count")
  args = parser.parse_args()

  logging.basicConfig(format='%(levelname)s:%(message)s', level=(logging.DEBUG if args.verbose>1 else logging.INFO))

  base_url = f"https://{args.api_host}{args.api_base_path}"

  fout = open(args.ofile, "w") if args.ofile else sys.stdout

  ids=[]
  if args.ifile:
    fin = open(args.ifile)
    while True:
      line = fin.readline()
      if not line: break
      ids.append(line.strip())
  elif args.ids:
    ids = re.split('[,\s]+', args.ids.strip())

  t0=time.time()

  if args.op =="list_contributors":
    bioregistry.ListEntities("contributors", base_url, fout)
  elif args.op =="list_collections":
    bioregistry.ListEntities("collections", base_url, fout)
  elif args.op =="list_contexts":
    bioregistry.ListEntities("contexts", base_url, fout)
  elif args.op =="list_metaregistry":
    bioregistry.ListEntities("metaregistry", base_url, fout)
  elif args.op =="list_registry":
    bioregistry.ListEntities("registry", base_url, fout)
  elif args.op =="get_reference":
    bioregistry.GetReference(ids, args.prefix, base_url, fout)

  else:
    parser.error("Invalid operation: {0}".format(args.op))

  logging.info(('Elapsed time: %s'%(time.strftime('%Hh:%Mm:%Ss', time.gmtime(time.time()-t0)))))



#!/usr/bin/env python3
"""
Utility functions for PDB REST API.
https://www.rcsb.org/docs/programmatic-access/web-services-overview
https://data.rcsb.org/redoc/

Major changes include all JSON output.
"""
###
import sys,os,re,time,json,logging,requests
#
API_HOST='data.rcsb.org'
API_BASE_PATH='/rest/v1'
API_BASE_URL = f"https://{API_HOST}{API_BASE_PATH}"
#
#############################################################################
def GetEntryData(eid, base_url=API_BASE_URL):
  """E.g. https://data.rcsb.org/rest/v1/core/entry/3ert"""
  response = requests.get(f"{base_url}/core/entry/{eid}")
  if response.status_code != 200:
    logging.error(f"status_code: {response.status_code}")
    return None
  logging.debug(response.content)
  result = response.json()
  return result

#############################################################################
def GetEntrys(eids, base_url=API_BASE_URL, fout=None):
  n_out=0; tags=[];
  for eid in eids:
    result = GetEntryData(eid, base_url)
    fout.write(json.dumps(result, indent=2)+"\n")
    n_out+=1
  logging.info(f"queries: {len(eids)}; entrys out: {n_out}")

#############################################################################
def GetChemicalData(cid, base_url):
  """Chemical component (ligands, small molecules and monomers) via CCD ID. E.g. https://data.rcsb.org/rest/v1/core/chemcomp/CFF"""
  response = requests.get(f"{base_url}/core/chemcomp/{cid}")
  if response.status_code != 200:
    logging.error(f"status_code: {response.status_code}")
    return None
  logging.debug(response.content)
  result = response.json()
  return result

#############################################################################
def GetChemicals(cids, druglike, base_url=API_BASE_URL, fout=None):
  n_all=0; n_out=0; n_rejected=0; tags=[];
  for cid in cids:
    result = GetChemicalData(base_url, cid)
    if result is None:
      continue
    n_all+=1
    if druglike and not ChemicalIsDruglike(result):
      n_rejected+=1
      continue
    fout.write(result)
    n_out+=1
  logging.info(f"queries: {len(cids)}; chemicals: {n_all}; chemicals out: {n_out}; rejected: {n_rejected}")

#############################################################################
def ChemicalIsDruglike(chem):
  '''Simple criteria to exclude monoatomic, polymers, etc.'''
  chem_comp = chem['chem_comp'] if 'chem_comp' in chem else None
  if not chem_comp: return False
  chemtype = chem_comp['type'] if 'type' in chem_comp else None
  if chemtype != 'non-polymer': return False
  mf = chem_comp['formula'] if 'formula' in chem_comp else None
  formula_weight = chem_comp['formula_weight'] if 'formula_weight' in chem_comp else None
  if not formula_weight or formula_weight<100 or formula_weight>1000: return False
  rcsb_chem_comp_descriptor = chem['rcsb_chem_comp_descriptor'] if 'rcsb_chem_comp_descriptor' in chem else None
  smi = rcsb_chem_comp_descriptor['smiles'] if 'smiles' in rcsb_chem_comp_descriptor else None
  if not smi: return False
  rcsb_chem_comp_info = chem['rcsb_chem_comp_info'] if 'rcsb_chem_comp_info' in chem else None
  atom_count = rcsb_chem_comp_info['atom_count'] if 'atom_count' in rcsb_chem_comp_info else None
  if atom_count<6: return False
  return True

#############################################################################
def ListEntrys(base_url=API_BASE_URL, fout=None):
  response = requests.get(f"{base_url}/holdings/current/entry_ids")
  if response.status_code != 200:
    logging.error(f"status_code: {response.status_code}")
    return []
  results = response.json()
  if fout is not None:
    for eid in results:
      fout.write(f"{eid}\n")
  logging.info(f"Entrys: {len(results)}")
  return results

#############################################################################
#############################################################################



#!/usr/bin/env python3
###
# https://gql.readthedocs.io/en/stable/
# https://gql.readthedocs.io/en/stable/usage/basic_usage.html
###
import sys,os,json,logging

from gql import Client, gql
from gql.transport.aiohttp import AIOHTTPTransport

#############################################################################
def RunQuery(graphql, base_url, fout):
  if graphql is None:
    logging.error("No query.")
    return

  transport = AIOHTTPTransport(url=base_url)
  client = Client(transport=transport, fetch_schema_from_transport=True)
  logging.debug(f"client.schema: '{client.schema}'")

  try:
    query = gql(graphql)
    result = client.execute(query)
    fout.write(json.dumps(result, indent=2)+"\n")
  except Exception as e:
    logging.error(e)

#############################################################################





#!/usr/bin/env python3
"""
Sparql endpoint client utility functions
"""
###
import os,sys,re,json,logging

import SPARQLWrapper

#############################################################################
def SparqlRequest(rq_code, rq_uri, defgraph=None, fmt=None):
  rq_results = None
  logging.debug(f"URI: {rq_uri}")
  logging.debug(rq_code)
  try:
    sparql = SPARQLWrapper.SPARQLWrapper(rq_uri)
    sparql.setQuery(rq_code)
    if fmt:
      sparql.setReturnFormat(fmt)
    rq_results = sparql.query()
  except Exception as e:
    logging.error(f"Error: {e}")
    errtype,val,traceback = sys.exc_info()
    logging.error(f"sys.exc_info:\n({errtype})\n{val}")
    if traceback: logging.info(f"traceback:\n{traceback}>")
    logging.error(str(rq_code))
  return rq_results

#############################################################################
def Results2TSV(results, variables, nmax, fout):
  if variables:
    variables = variables.strip()
    variables = re.split(r'[\s,]+', variables)
  else:
    variables=[]
  if not variables:
    variables = results["head"]["vars"]
  logging.debug("selected vars: "+(','.join(variables)))
  if "link" in results["head"]:
    logging.debug("links: "+(','.join(results['head']['link'])))
  if "distinct" in results["results"]:
    logging.debug(f"distinct: {results['results']['distinct']}")
  if "ordered" in results["results"]:
    logging.debug(f"ordered: {results['results']['ordered']}")
  n_row=0;
  fout.write(('\t'.join(variables))+"\n")
  for binding in results["results"]["bindings"]:
    logging.debug(json.dumps(binding, indent=2))
    row_out=[]
    for var in variables:
      if var in binding:
        val = binding[var]["value"]
        if binding[var]["type"]=='uri':
          val=(f"<{val}>")
      else:
        val=''
      try:
        val=val.encode('utf8')
      except:
        logging.info(f"unicode error [{n_row}]")
        val='error'
      row_out.append(val)
    fout.write(('\t'.join(row_out))+"\n")
    n_row+=1
    if n_row==nmax:
      logging.info(f"Output truncated at NMAX = {nmax}.")
      break
  logging.info(f"N = {n_row}")

#############################################################################
### Return list of rows, each row list of values.
#############################################################################
def Results2List(results, variables, nmax):
  if variables:
    variables = variables.strip()
    variables = re.split(r'[\s,]+', variables)
  else:
    variables=[]
  if not variables:
    variables = results["head"]["vars"]
  logging.debug("selected vars: "+(','.join(variables)))
  if "link" in results["head"]:
    logging.debug("links: "+(','.join(results['head']['link'])))
  if "distinct" in results["results"]:
    logging.debug(f"distinct: {results['results']['distinct']}")
  if "ordered" in results["results"]:
    logging.debug(f"ordered: {results['results']['ordered']}")
  n_row=0
  results_out=[]
  for binding in results['results']['bindings']:
    logging.debug(json.dumps(binding, indent=2))
    row_out=[]
    for var in variables:
      if var in binding:
        val = binding[var]['value']
        if binding[var]['type']=='uri':
          val=(f"<{val}>")
      else:
        val=''
      try:
        val=val.encode('utf8')
      except:
        logging.info(f"unicode error [{n_row}]")
        val='error'
      row_out.append(val)
    results_out.append(row_out)
    n_row+=1
    if n_row==nmax:
      logging.info(f"NOTE: Output truncated at NMAX = {nmax}.")
      break
  logging.info(f"N = {n_row}")
  return results_out

#############################################################################
#  FILTER (lang(?propval) = "en" || lang(?propval) = "" ) ##removes depiction
#  FILTER (lang(?propval) IN ( "en" , "" )) ##removes depiction
def Drugname2Sparql(drugname):
  rq_code = f"""\
SELECT DISTINCT
  ?drug ?propname ?propval
WHERE
{{
  {{
    <http://dbpedia.org/resource/{drugname}> rdf:type <http://dbpedia.org/ontology/Drug> .
    <http://dbpedia.org/resource/{drugname}> ?propname ?propval .
  }}
  UNION
  {{
    ?drug rdf:type <http://dbpedia.org/ontology/Drug> .
    ?drug ?propname ?propval .
    {{ ?drug <http://dbpedia.org/property/licenceEu> ?synonym . }}
    UNION
    {{ ?drug <http://dbpedia.org/property/licenceUs> ?synonym . }}
    UNION
    {{ ?drug <http://dbpedia.org/property/tradename> ?synonym . }}
    UNION
    {{
      ?drug rdfs:label ?synonym .
      FILTER (regex(?synonym, "^{drugname}$", "i"))
    }}
    FILTER (regex(?synonym, "^{drugname}$", "i"))
  }}
}}
"""
  return rq_code

#############################################################################
def Test(drugname="metformin", fmt=SPARQLWrapper.JSON):
  RQ_URI = 'http://dbpedia.org/sparql'
  logging.info(f"drugname: '{drugname}'; endpoint: {RQ_URI}")
  rq_code = Drugname2Sparql(drugname)
  logging.debug(f"sparql:\n{rq_code}")
  rq_results = SparqlRequest(rq_code, RQ_URI, defgraph=None, fmt=fmt)
  if fmt in (SPARQLWrapper.JSON, SPARQLWrapper.JSONLD):
    print(json.dumps(rq_results.convert(), indent=2))
  elif fmt in (SPARQLWrapper.XML, SPARQLWrapper.RDFXML, SPARQLWrapper.RDF):
    print(rq_results.convert().toxml())
  else:
    print(rq_results.convert().decode("utf-8"))

#############################################################################



#!/usr/bin/env python3
"""
Pharos GraphQL API client
https://pharos.nih.gov/api
"""
###
import sys,os,argparse,json,re,time,logging
#
import gql
import python_graphql_client
#
from ...util import graphql as util_graphql
from ...idg import pharos
#
#############################################################################
def DemoGql():
  return """\
{
  search(term: "gout") {
    diseaseResult {
      count
      facets {
        facet
        values {
          name
          value
        }
      }
      diseases {
        name
        associationCount
        associations {
          did
          name
          type
        }
      }
    }
  }
}
"""

#############################################################################
if __name__=='__main__':
  parser = argparse.ArgumentParser(description='Pharos GraphQL API client')
  ops = [
	'get_targets', 'get_diseases', 'test',
	'gql_query', 'gql_demo', ]
  parser.add_argument("op", choices=ops, help='OPERATION')
  parser.add_argument("--i", dest="ifile", help="input file, target IDs")
  parser.add_argument("--i_gql", dest="ifile_gql", help="input file, GraphQL")
  parser.add_argument("--graphql", help="input GraphQL")
  parser.add_argument("--ids", dest="ids", help="IDs, target, comma-separated")
  parser.add_argument("--o", dest="ofile", help="output (TSV)")
  parser.add_argument("--idtype_target", choices=list(pharos.IDTYPES_TARGET.keys()), default='sym', help="target ID type")
  parser.add_argument("--idtype_disease", choices=list(pharos.IDTYPES_DISEASE.keys()), default='name', help="disease ID type")
  parser.add_argument("--nmax", type=int, help="max to return")
  parser.add_argument("--api_host", default=pharos.API_HOST)
  parser.add_argument("--api_base_path", default=pharos.API_BASE_PATH)
  parser.add_argument("-v", "--verbose", default=0, action="count")
  args = parser.parse_args()

  logging.basicConfig(format='%(levelname)s:%(message)s', level=(logging.DEBUG if args.verbose>1 else logging.INFO))

  base_url = "https://"+args.api_host+args.api_base_path

  fout = open(args.ofile, "w+") if args.ofile else sys.stdout

  logging.debug(f"gql version: {gql.__version__}")

  ids=[];
  if args.ifile:
    fin = open(args.ifile)
    while True:
      line = fin.readline()
      if not line: break
      ids.append(line.rstrip())
    fin.close()
  elif args.ids:
    ids = re.split(r'\s*,\s*', args.ids.strip())
  logging.info(f"Input IDs: {len(ids)}")

  graphql=None;
  if args.ifile_gql:
    with open(args.ifile_gql) as fin:
      graphql = fin.read()
  elif args.graphql:
    graphql = args.graphql
  logging.debug(graphql)

  t0 = time.time()

  if re.match(r'^get_', args.op) and not ids:
    parser.error(f"{args.op} requires IDs.")

  if args.op=='get_targets':
    pharos.GetTargets(ids, args.idtype_target, base_url, fout)

  elif args.op=='get_diseases':
    pharos.GetDiseases(ids, args.idtype_disease, base_url, fout)

  elif args.op=='test':
    pharos.Test(base_url, fout)

  elif args.op == 'gql_query':
    util_graphql.RunQuery(graphql, base_url, fout)

  elif args.op == 'gql_demo':
    util_graphql.RunQuery(DemoGql(), base_url, fout)

  else:
    logging.error(f"Invalid operation: {args.op}")

  logging.info("Elapsed time: "+time.strftime('%Hh:%Mm:%Ss', time.gmtime(time.time()-t0)))



#!/usr/bin/env python3
"""
"""
import sys,os,re,time,json,logging,requests,tqdm
import urllib,urllib.parse
import pandas as pd
#
API_HOST='gsrs.ncats.nih.gov'
API_BASE_PATH='/ginas/app/api/v1'
API_BASE_URL='https://'+API_HOST+API_BASE_PATH
#
NCHUNK=100
#
##############################################################################
def ListVocabularies(base_url=API_BASE_URL, fout=None):
  n_out=0; tags=None; tags_term=None; df=None;
  size=NCHUNK; skip=0;
  while True:
    url_this = f"{base_url}/vocabularies?skip={skip}&top={size}"
    response = requests.get(url_this)
    logging.debug(response.text)
    if response.status_code!=200: break
    results = response.json()
    things = results['content']
    if not things: break
    for thing in things:
      if not tags:
        tags = list(thing.keys())
        for tag in tags[:]:
          if type(thing[tag]) in (list, dict):
            logging.info(f"Ignoring field: {tag}")
            tags.remove(tag)
      for term in thing['terms']:
        if not tags_term:
          tags_term = list(term.keys())
        df_this = pd.DataFrame({tag:[thing[tag] if tag in thing else ''] for tag in tags})
        df_this = pd.concat([df_this, pd.DataFrame({f"term_{tag}":[term[tag] if tag in term else ''] for tag in tags_term})], axis=1)
        if fout is None: df = pd.concat([df, df_this])
        else: df_this.to_csv(fout, sep="\t", index=False, header=bool(n_out==0))
        n_out += df_this.shape[0]
    if results['count']<NCHUNK: break
    skip+=NCHUNK
  logging.info(f"n_out: {n_out}")
  return df

##############################################################################
def ListSubstances(base_url=API_BASE_URL, fout=None):
  n_out=0; tags=None; df=None; tq=None;
  size=NCHUNK; skip=0;
  while True:
    url_this = f"{base_url}/substances?skip={skip}&top={size}"
    response = requests.get(url_this)
    logging.debug(response.text)
    if response.status_code!=200: break
    results = response.json()
    if not tq: tq = tqdm.tqdm(total=results['total'])
    things = results['content']
    if not things: break
    for thing in things:
      if not tags:
        tags = list(thing.keys())
        for tag in tags[:]:
          if type(thing[tag]) in (list, dict):
            logging.info(f"Ignoring field: {tag}")
            tags.remove(tag)
      df_this = pd.DataFrame({tag:[thing[tag] if tag in thing else ''] for tag in tags})
      if fout is None: df = pd.concat([df, df_this])
      else: df_this.to_csv(fout, sep="\t", index=False, header=bool(n_out==0))
      n_out += df_this.shape[0]
    tq.update(n=results['count'])
    if results['count']<NCHUNK: break
    skip+=NCHUNK
  if tq is not None: tq.close()
  logging.info(f"n_out: {n_out}")
  return df

#############################################################################
def Search(query, base_url=API_BASE_URL, fout=None):
  n_out=0; tags=None; df=None;
  size=NCHUNK; skip=0;
  url_this = f"{base_url}/substances/search?q={urllib.parse.quote(query)}&top={size}"
  while True:
    response = requests.get(url_this)
    logging.debug(response.text)
    if response.status_code!=200: break
    results = response.json()
    things = results['content']
    if not things: break
    for thing in things:
      if not tags:
        tags = list(thing.keys())
        for tag in tags[:]:
          if type(thing[tag]) in (list, dict):
            logging.info(f"Ignoring field: {tag}")
            tags.remove(tag)
      df_this = pd.DataFrame({tag:[thing[tag] if tag in thing else ''] for tag in tags})
      df_this = pd.concat([df_this,
pd.DataFrame({
	"struct_id":[thing['structure']['id'] if 'structure' in thing and 'id' in thing['structure'] else ''],
	"struct_smiles":[thing['structure']['smiles'] if 'structure' in thing and 'smiles' in thing['structure'] else '']})], axis=1)
      if fout is None: df = pd.concat([df, df_this])
      else: df_this.to_csv(fout, sep="\t", index=False, header=bool(n_out==0))
      n_out += df_this.shape[0]
    if results['count']<NCHUNK: break
    url_this = results['nextPageUri']
  logging.info(f"n_out: {n_out}")
  return df

#############################################################################
def GetSubstance(ids, base_url=API_BASE_URL, fout=None):
  n_out=0; tags=None; df=None;
  for id_this in ids:
    url_this = f"{base_url}/substances({id_this})"
    response = requests.get(url_this)
    logging.debug(response.text)
    if response.status_code!=200: continue
    thing = response.json()
    if not tags:
      tags = list(thing.keys())
      for tag in tags[:]:
        if type(thing[tag]) in (list, dict):
          logging.info(f"Ignoring field: {tag}")
          tags.remove(tag)
    df_this = pd.DataFrame({tag:[thing[tag] if tag in thing else ''] for tag in tags})
    df_this = pd.concat([df_this,
	pd.DataFrame({
	"struct_id":[thing['structure']['id'] if 'structure' in thing and 'id' in thing['structure'] else ''],
	"struct_smiles":[thing['structure']['smiles'] if 'structure' in thing and 'smiles' in thing['structure'] else '']})], axis=1)
    if fout is None: df = pd.concat([df, df_this])
    else: df_this.to_csv(fout, sep="\t", index=False, header=bool(n_out==0))
    n_out += df_this.shape[0]
  logging.info(f"n_out: {n_out}")
  return df

#############################################################################
def GetSubstanceNames(ids, base_url=API_BASE_URL, fout=None):
  n_out=0; tags=None; df=None;
  for id_this in ids:
    url_this = f"{base_url}/substances({id_this})/names"
    response = requests.get(url_this)
    logging.debug(response.text)
    if response.status_code!=200: continue
    names = response.json()
    for name in names:
      if not tags:
        tags = list(name.keys())
        for tag in tags[:]:
          if type(name[tag]) in (list, dict):
            logging.info(f"Ignoring field: {tag}")
            tags.remove(tag)
      df_this = pd.DataFrame({tag:[name[tag] if tag in name else ''] for tag in tags})
      if fout is None: df = pd.concat([df, df_this])
      else: df_this.to_csv(fout, sep="\t", index=False, header=bool(n_out==0))
      n_out += df_this.shape[0]
  logging.info(f"n_out: {n_out}")
  return df

#############################################################################



#!/usr/bin/env python3
'''
https://www.ncbi.nlm.nih.gov/pmc/tools/developers/
https://www.ncbi.nlm.nih.gov/pmc/tools/get-metadata/
'''
import os,sys,io,re,json,time,requests,urllib.parse,logging,tqdm,tqdm.auto
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry
import pandas as pd

from xml.etree import ElementTree

from .. import util
from ..util import xml as util_xml

API_HOST="eutils.ncbi.nlm.nih.gov"
API_BASE_PATH="/entrez/eutils"
BASE_URL=f"https://{API_HOST}{API_BASE_PATH}"

#############################################################################
def GetESummary(ids, skip=0, nmax=None, base_url=BASE_URL, fout=None):
  n_out=0; n_err=0; tq=None; tags=None; df=None; i_this=0;
  if skip: logging.debug(f"skip: [1-{skip}]")
  for id_this in ids:
    i_this+=1
    if i_this<skip: continue
    if tq is None: tq = tqdm.tqdm(total=(len(ids)-skip if nmax is None else nmax))
    url_this = f"{base_url}/esummary.fcgi?db=pubmed&id={id_this}&retmode=json"
    try:
      response = requests.get(url_this)
    except Exception as e:
      logging.error(f"{e}")
      n_err+=1
      continue
    if response.status_code!=200:
      logging.debug(f"Status code: {response.status_code}")
      continue
    result = response.json()
    logging.debug(json.dumps(result, indent=2))
    pubs = result["result"] if "result" in result else []
    uids = pubs["uids"] if "uids" in pubs else []
    for uid in uids:
      pub = pubs[uid] if uid in pubs else None
      if not pub:
        logging.error(f"{i_this}. PMID not found: {id_this}")
        continue #Error
      if tags is None:
        tags = [tag for tag in pub.keys() if type(pub[tag]) not in (list, dict)]
      df_this = pd.DataFrame({tag:[pub[tag]] for tag in tags})
      if fout is not None:
        df_this.to_csv(fout, sep="\t", index=False, header=bool(n_out==0))
      else:
        df = pd.concat([df, df_this])
    n_out+=1
    tq.update(n=1)
    if nmax is not None and i_this-skip>nmax: break
  tq.close()
  logging.info(f"n_out: {n_out}; n_err: {n_err}")
  return df

#############################################################################
def GetRecord(ids, skip=0, nmax=None, base_url=BASE_URL, fout=None):
  """Only get title, abstract, journal, year.  Must process XML since no retmode=json supported."""
  n_out=0; n_err=0; tq=None; tags=None; df=None; i_this=0;
  if skip: logging.debug(f"skip: [1-{skip}]")
  for id_this in ids:
    i_this+=1
    if i_this<skip: continue
    if tq is None: tq = tqdm.tqdm(total=min(len(ids)-skip, nmax if nmax is not None else float("inf")))
    url_this = f"{base_url}/efetch.fcgi?db=pubmed&id={id_this}"
    try:
      response = requests.get(url_this)
    except Exception as e:
      logging.error(f"{e}")
      n_err+=1
      continue
    if response.status_code!=200:
      logging.debug(f"Status code: {response.status_code}")
      continue
    pubxml = response.content.decode('utf-8')
    logging.debug(pubxml)
    try:
      elt = ElementTree.parse(io.StringIO(pubxml))
    except Exception as e:
      logging.error(f"XML parse error: {e}; xml={pubxml}")
      n_err+=1
      continue
    logging.debug(f"XML parsed ok.")
    articles = elt.iterfind("PubmedArticle")
    for article in articles:
      logging.debug(f"article.tag <{article.tag}>")
      title = util_xml.GetFirstLeafValByTagName(article, "ArticleTitle") 
      abstracttext = util_xml.GetFirstLeafValByTagName(article, "AbstractText") 
      journal = util_xml.GetFirstLeafValByTagName(article, "Title") 
      year = util_xml.GetFirstLeafValByTagName(article, "Year") 
      pmid = util_xml.GetFirstLeafValByTagName(article, "PMID") 
      if pmid != id_this:
        logging.error(f"PMID {pmid} != {id_this}")
        n_err+=1
        continue
      authorlastname=None;
      #authorlist = article.find("AuthorList") #Why not working?
      if article.iter("AuthorList") is not None and len(list(article.iter("AuthorList")))>0: #Kludge
        authorlist = list(article.iter("AuthorList"))[0] #Kludge
        authors = authorlist.findall("Author")
        for author in authors:
          authorlastname = util_xml.GetFirstLeafValByTagName(author, "LastName")
          break
      df_this = pd.DataFrame({
	"pmid":[pmid],
	"title":[title],
	"abstract":[abstracttext],
	"author":[authorlastname],
	"journal":[journal],
	"year":[year]
	})
      if fout is not None:
        df_this.to_csv(fout, sep="\t", index=False, header=bool(n_out==0))
      else:
        df = pd.concat([df, df_this])
      n_out+=df_this.shape[0]
    tq.update(n=1)
  tq.close()
  logging.info(f"n_ids: {len(ids)}; n_out: {n_out}; n_err: {n_err}")
  return df

#############################################################################



#!/usr/bin/env python3
"""
Sparql endpoint client
"""
#############################################################################
import os,sys,json,argparse,re,time,logging

import SPARQLWrapper

from .. import sparql as util_sparql

FMTS={
	'JSON':SPARQLWrapper.Wrapper.JSON,
	'JSONLD':SPARQLWrapper.Wrapper.JSONLD,
	'XML':SPARQLWrapper.Wrapper.XML,
	'RDF':SPARQLWrapper.Wrapper.RDF,
	'RDFXML':SPARQLWrapper.Wrapper.RDFXML,
	'N3':SPARQLWrapper.Wrapper.N3,
	'TTL':SPARQLWrapper.Wrapper.TURTLE,
	'CSV':SPARQLWrapper.Wrapper.CSV,
	'TSV':SPARQLWrapper.Wrapper.TSV
	}

#############################################################################
if __name__=='__main__':
  epilog = f"SPARQLWRAPPERVERSION: {SPARQLWrapper.__version__}"
  parser = argparse.ArgumentParser(description='Sparql utilities', epilog=epilog)
  ENDPOINT="http://dbpedia.org/sparql"
  OPERATIONS = ["query", "test", ]
  parser.add_argument("op", help="OPERATION")
  parser.add_argument("--rqfile", help="Sparql input file")
  parser.add_argument("--rq", help="Sparql input string")
  parser.add_argument("--o", dest="ofile", help="output results (TSV)")
  parser.add_argument("--endpoint", default=ENDPOINT, help=f"Sparql endpoint [{ENDPOINT}]")
  parser.add_argument("--defgraph", help="default graph URL")
  parser.add_argument("--nmax", type=int, default=1000, help="max returned triples")
  parser.add_argument("--fmt", choices=FMTS.keys(), default='JSON', help="output format")
  parser.add_argument("--test_drugname", default="metformin", help="test drugname query")
  parser.add_argument("-v", "--verbose", action="count", default=0)
  args = parser.parse_args()

  logging.basicConfig(format='%(levelname)s:%(message)s', level=(logging.DEBUG if args.verbose>1 else logging.INFO))

  t0=time.time()
  fout = open(args.ofile, 'w+') if args.ofile else sys.stdout

  logging.debug(f"SPARQLWrapper_Version: {SPARQLWrapper.__version__}")
  logging.debug(f"fmt: {args.fmt}")

  if args.op== "test":
    util_sparql.Test(args.test_drugname, FMTS[args.fmt])

  elif args.op== "query":
    if args.rqfile:
      with open(args.rqfile) as fin:
        rqcode = fin.read()
    elif args.rq:
      rqcode = args.rq
    else:
      parser.error("--rq or --rqfile required.")

    if args.fmt=='XML':
      results = util_sparql.SparqlRequest(rqcode, args.endpoint, args.defgraph, FMTS[args.fmt])
      results = results.convert()
      fout.write(f"{results.toxml()}\n")
    elif args.fmt=='JSON':
      results = util_sparql.SparqlRequest(rqcode, args.endpoint, args.defgraph, FMTS[args.fmt])
      results = results.convert()
      fout.write(json.dumps(results, indent=2)+"\n")
    elif args.fmt=='N3':
      results = util_sparql.SparqlRequest(rqcode, args.endpoint, args.defgraph, FMTS[args.fmt])
      fout.write(results._convertN3()+"\n")
    elif args.fmt=='TSV':
      results = util_sparql.SparqlRequest(rqcode, args.endpoint, args.defgraph, FMTS['JSON'])
      results = results.convert()
      util_sparql.Results2TSV(results, [], nmax, fout)
    else:
      results = util_sparql.SparqlRequest(rqcode, args.endpoint, args.defgraph, None)
      fout.write(str(results)+"\n")

  else:
    parser.error(f"Unsupported operation: {args.op}")



#!/usr/bin/env python3
"""
Utility functions for UBKG REST API.
See: https://smart-api.info/ui/55be2831c69b17f6bc975ddb58cabb5e
C0006142,C0678222
"""
import sys,os,re,requests,urllib,json,time,logging,tqdm
import pandas as pd

#
API_HOST="datadistillery.api.sennetconsortium.org"
API_BASE_PATH=""
BASE_URL="https://"+API_HOST+API_BASE_PATH
#
HEADERS={"Accept":"application/json"}
#
##############################################################################
def Info(api_key, base_url=BASE_URL, fout=None):
  headers = {**HEADERS, **{"X-API-KEY": api_key}}
  response = requests.get(base_url+"/database/server", headers=headers)
  result = response.json()
  logging.debug(json.dumps(result, sort_keys=True, indent=2))
  df = pd.DataFrame({"field":result.keys(), "value":result.values()})
  if fout: df.to_csv(fout, sep="\t", index=False)
  return df

##############################################################################
def ListSABs(api_key, base_url=BASE_URL, fout=None):
  headers = {**HEADERS, **{"X-API-KEY": api_key}}
  response = requests.get(base_url+"/sabs", headers=headers)
  result = response.json()
  logging.debug(json.dumps(result, sort_keys=True, indent=2))
  df = pd.DataFrame(result)
  if fout: df.to_csv(fout, sep="\t", index=False)
  logging.info(f"SABs: {df.shape[0]}")
  return df

##############################################################################
def ListNodeTypes(api_key, base_url=BASE_URL, fout=None):
  headers = {**HEADERS, **{"X-API-KEY": api_key}}
  response = requests.get(base_url+"/node-types", headers=headers)
  result = response.json()
  logging.debug(json.dumps(result, sort_keys=True, indent=2))
  df = pd.DataFrame(result)
  if fout: df.to_csv(fout, sep="\t", index=False)
  logging.info(f"Node-types: {df.shape[0]}")
  return df

##############################################################################
def ListNodeTypeCounts(api_key, base_url=BASE_URL, fout=None):
  headers = {**HEADERS, **{"X-API-KEY": api_key}}
  node_types = ListNodeTypes(api_key, base_url)["node_types"]
  n_type=0; n_out=0; df=None;
  for node_type in node_types:
    n_type+=1
    response = requests.get(f"{base_url}/node-types/{node_type}/counts", headers=headers)
    result = response.json()
    logging.debug(json.dumps(result, sort_keys=True, indent=2))
    df_this = pd.DataFrame(result)
    df_this.insert(loc=0, column='node_type', value=node_type)
    df_this = df_this[["node_type", "total_count"]]
    if fout: df_this.to_csv(fout, sep="\t", index=False, header=bool(n_out==0))
    else: df = pd.concat([df, df_this])
    n_out += 1
  return df

##############################################################################
def ListRelationshipTypes(api_key, base_url=BASE_URL, fout=None):
  headers = {**HEADERS, **{"X-API-KEY": api_key}}
  response = requests.get(base_url+"/relationship-types", headers=headers)
  result = response.json()
  logging.debug(json.dumps(result, sort_keys=True, indent=2))
  df = pd.DataFrame(result)
  if fout: df.to_csv(fout, sep="\t", index=False)
  logging.info(f"Relationship-types: {df.shape[0]}")
  return df

##############################################################################
def ListPropertyTypes(api_key, base_url=BASE_URL, fout=None):
  headers = {**HEADERS, **{"X-API-KEY": api_key}}
  response = requests.get(base_url+"/property-types", headers=headers)
  result = response.json()
  logging.debug(json.dumps(result, sort_keys=True, indent=2))
  df = pd.DataFrame(result)
  if fout: df.to_csv(fout, sep="\t", index=False)
  logging.info(f"Property-types: {df.shape[0]}")
  return df

##############################################################################
def ListSemanticTypes(api_key, base_url=BASE_URL, fout=None):
  headers = {**HEADERS, **{"X-API-KEY": api_key}}
  response = requests.get(base_url+"/semantics/semantic-types", headers=headers)
  result = response.json()
  logging.debug(json.dumps(result, sort_keys=True, indent=2))
  semantic_types = result["semantic_types"]
  df = pd.DataFrame.from_records(semantic_types)
  if fout: df.to_csv(fout, sep="\t", index=False)
  logging.info(f"Semantic-types: {df_this.shape[0]}")
  return df

##############################################################################
def GetConcept2Codes(ids, api_key, base_url=BASE_URL, fout=None):
  headers = {**HEADERS, **{"X-API-KEY": api_key}}
  n_in=0; n_out=0; n_code=0; df=None;
  for id_this in ids:
    n_in += 1
    response = requests.get(f"{base_url}/concepts/{id_this}/codes", headers=headers)
    if not response:
      continue
    result = response.json()
    logging.debug(json.dumps(result, sort_keys=True, indent=2))
    codes = result
    if not codes or len(codes)==0: continue
    df_this = pd.DataFrame({"concept":[id_this for i in range(len(codes))], "code":codes})
    if fout: df_this.to_csv(fout, sep="\t", index=False, header=bool(n_out==0))
    else: df = pd.concat([df, df_this])
    n_out += 1
    n_code += len(codes)
  logging.info(f"Concepts in: {n_in}; out: {n_out}; codes out: {n_code}")
  return df

##############################################################################
def GetConcept2Concepts(ids, api_key, base_url=BASE_URL, fout=None):
  headers = {**HEADERS, **{"X-API-KEY": api_key}}
  n_in=0; n_out=0; n_concept=0; df=None;
  for id_this in ids:
    n_in += 1
    response = requests.get(f"{base_url}/concepts/{id_this}/concepts", headers=headers)
    if not response:
      continue
    result = response.json()
    logging.debug(json.dumps(result, sort_keys=True, indent=2))
    concepts = result
    if not concepts or len(concepts)==0: continue
    df_this = pd.DataFrame.from_records(concepts)
    df_this.columns = ["conceptB", "preftermB", "relationship", "sabB"]
    df_this.insert(loc=0, column='conceptA', value=id_this)
    if fout: df_this.to_csv(fout, sep="\t", index=False, header=bool(n_out==0))
    else: df = pd.concat([df, df_this])
    n_out += 1
    n_concept += len(concepts)
  logging.info(f"Concepts in: {n_in}; out: {n_out}; concepts out: {n_concept}")
  return df

##############################################################################
def GetConcept2Definitions(ids, api_key, base_url=BASE_URL, fout=None):
  headers = {**HEADERS, **{"X-API-KEY": api_key}}
  n_in=0; n_out=0; n_def=0; df=None;
  for id_this in ids:
    n_in += 1
    response = requests.get(f"{base_url}/concepts/{id_this}/definitions", headers=headers)
    if not response:
      continue
    result = response.json()
    logging.debug(json.dumps(result, sort_keys=True, indent=2))
    defs = result
    if not defs or len(defs)==0: continue
    df_this = pd.DataFrame.from_records(defs)
    df_this = df_this[["sab", "definition"]]
    df_this.insert(loc=0, column='concept', value=id_this)
    if fout: df_this.to_csv(fout, sep="\t", index=False, header=bool(n_out==0))
    else: df = pd.concat([df, df_this])
    n_out += 1
    n_def += len(defs)
  logging.info(f"Concepts in: {n_in}; out: {n_out}; defs out: {n_def}")
  return df

##############################################################################
def GetConcept2Paths(ids, sab, rel, mindepth, maxdepth, api_key, base_url=BASE_URL, fout=None):
  '''JSON output'''
  headers = {**HEADERS, **{"X-API-KEY": api_key}}
  n_in=0; n_node_out=0; n_edge_out=0;
  for id_this in ids:
    n_in += 1
    response = requests.get(f"{base_url}/concepts/{id_this}/paths/expand?sab={sab}&rel={rel}&mindepth={mindepth}&maxdepth={maxdepth}&skip=0&limit=10", headers=headers)
    if not response:
      continue
    result = response.json()
    n_edge_out += len(result["edges"])
    n_node_out += len(result["nodes"])
    logging.debug(json.dumps(result, sort_keys=True, indent=2))
    if fout: fout.write(json.dumps(result, sort_keys=True, indent=2))
  logging.info(f"Concepts in: {n_in}; nodes out: {n_node_out}; edges out: {n_edge_out}")

##############################################################################
def GetConcept2Trees(ids, sab, rel, mindepth, maxdepth, api_key, base_url=BASE_URL, fout=None):
  '''JSON output'''
  headers = {**HEADERS, **{"X-API-KEY": api_key}}
  n_in=0; n_node_out=0; n_edge_out=0;
  for id_this in ids:
    n_in += 1
    response = requests.get(f"{base_url}/concepts/{id_this}/paths/trees?sab={sab}&rel={rel}&mindepth={mindepth}&maxdepth={maxdepth}&skip=0&limit=10", headers=headers)
    if not response:
      continue
    result = response.json()
    n_edge_out += len(result["edges"])
    n_node_out += len(result["nodes"])
    logging.debug(json.dumps(result, sort_keys=True, indent=2))
    if fout: fout.write(json.dumps(result, sort_keys=True, indent=2))
  logging.info(f"Concepts in: {n_in}; nodes out: {n_node_out}; edges out: {n_edge_out}")

##############################################################################
def GetTerm2Concepts(term, api_key, base_url=BASE_URL, fout=None):
  headers = {**HEADERS, **{"X-API-KEY": api_key}}
  response = requests.get("{}/terms/{}/concepts".format(base_url, urllib.parse.quote(term)), headers=headers)
  if not response:
    logging.info(f"Not found: {term}")
    return
  result = response.json()
  logging.debug(json.dumps(result, sort_keys=True, indent=2))
  concepts = result
  if not concepts or len(concepts)==0:
    logging.info(f"Not found: {term}")
    return
  df = pd.DataFrame({"term":[term for i in range(len(concepts))], "concept":concepts})
  if fout: df.to_csv(fout, sep="\t", index=False)
  logging.info(f"Concepts out: {df.shape[0]}")
  return df




#!/usr/bin/env python3
"""
See: https://oncotree.mskcc.org/
See: https://oncotree.mskcc.org/#/home
"""
###
import sys,os,re,argparse,time,json,logging
import pandas as pd
#
from .. import oncotree
#
##############################################################################
if __name__=='__main__':
  QUERY_TYPES = ["umls", "name", "nci", "level", "code", "mainType"]
  epilog='''\
'''
  parser = argparse.ArgumentParser(description='Oncotree REST API client', epilog=epilog)
  ops = ['search', 'get', 'info', 'list_versions', 'list_main_types', 'list_tumor_types']
  parser.add_argument("op", choices=ops, help='OPERATION')
  parser.add_argument("--o", dest="ofile", help="output (TSV)")
  parser.add_argument("--i", dest="ifile", help="ID file")
  parser.add_argument("--ids", help="IDs, comma-separated")
  parser.add_argument("--query", help="search query")
  parser.add_argument("--query_type", choices=QUERY_TYPES, help=f"query type, one of: {QUERY_TYPES}")
  parser.add_argument("--exact_match", action="store_true", help="search criterion")
  parser.add_argument("--query_levels", default="2,3,4,5", help=f"query levels, 1-5, comma-separated list")
  parser.add_argument("--nmax", type=int, default=None, help="max records")
  parser.add_argument("--skip", type=int, default=0, help="skip 1st SKIP queries")
  parser.add_argument("--api_host", default=oncotree.API_HOST)
  parser.add_argument("--api_base_path", default=oncotree.API_BASE_PATH)
  parser.add_argument("-v", "--verbose", default=0, action="count")
  args = parser.parse_args()

  logging.basicConfig(format='%(levelname)s:%(message)s', level=(logging.DEBUG if args.verbose>1 else logging.INFO))

  BASE_URL='http://'+args.api_host+args.api_base_path

  fout = open(args.ofile,"w") if args.ofile else sys.stdout

  t0=time.time()

  ids=[];
  if args.ifile:
    with open(args.ifile) as fin:
      while True:
        line = fin.readline()
        if not line: break
        if line.rstrip(): ids.append(line.rstrip())
  elif args.ids:
    ids = re.split('[, ]+', args.ids.strip())

  if args.op == "info":
    oncotree.Info(BASE_URL, fout)

  elif args.op == "get":
    oncotree.Get(ids, args.skip, BASE_URL, fout)

  elif args.op == "list_versions":
    oncotree.ListVersions(BASE_URL, fout)

  elif args.op == "list_main_types":
    oncotree.ListMainTypes(BASE_URL, fout)

  elif args.op == "list_tumor_types":
    oncotree.ListTumorTypes(BASE_URL, fout)

  elif args.op == "search":
    oncotree.SearchTumorTypes(args.query, args.query_type, args.exact_match, args.query_levels, BASE_URL, fout)

  else:
    parser.error(f"Invalid operation: {args.op}")

  logging.info(('elapsed time: %s'%(time.strftime('%Hh:%Mm:%Ss', time.gmtime(time.time()-t0)))))



#!/usr/bin/env python3
'''
Parse, process Entrez PubMed XML (summaries or full), normally obtained via Entrez eUtils,
e.g. entrez_pubmed_search.pl.

Note that other Entrez XML (e.g. PubChem) very similar.
'''
import os,sys,re,time,argparse,logging

from xml.etree import ElementTree
from xml.parsers import expat

from .. import util
from ..pubmed import Utils_XML

#############################################################################
if __name__=='__main__':
  parser = argparse.ArgumentParser(description='process PubMed XML (summaries or full), typically obtained via Entrez eUtils.')
  ops=['summary2tsv', 'summary2abstract', 'full2tsv', 'full2abstract', 'full2authorlist']
  parser.add_argument("op", choices=ops, help="OPERATION")
  parser.add_argument("--i", dest="ifile", required=True, help="input file, XML")
  parser.add_argument("--ids", help="PubMed IDs, comma-separated (ex:25533513)")
  parser.add_argument("--idfile", help="input file, PubMed IDs")
  parser.add_argument("--nmax", help="max to return")
  parser.add_argument("--o", dest="ofile", help="output (TSV)")
  parser.add_argument("--odir", help="output directory")
  parser.add_argument("-v", "--verbose", default=0, action="count")
  args = parser.parse_args()

  logging.basicConfig(format='%(levelname)s:%(message)s', level=(logging.DEBUG if args.verbose>1 else logging.INFO))

  fin = open(args.ifile)

  pmids=[]
  if args.idfile:
    fin_ids = open(args.idfile)
    if not fin_ids: logging.error(f"Cannot open idfile: {args.idfile}")
    while True:
      line = fin_ids.readline()
      if not line: break
      try:
        ids.append(int(line.rstrip()))
      except:
        logging.error(f"Bad input ID: {line}")
        continue
    if args.verbose:
      logging.info(f"Input IDs: {len(ids)}")
    fin_ids.close()

  fout = open(args.ofile, "w") if args.ofile else sys.stdout

  try:
    tree = ElementTree.parse(fin)
  except Exception as e:
    logging.error(f"Failed to parse XML: {e}")

  if args.op == 'summary2tsv':
    Utils_XML.PubMed_Docsum2Tsv(tree, fout)

  elif args.op == 'full2tsv':
    Utils_XML.PubMed_Doc2Tsv(tree, fout)

  elif args.op == 'full2authorlist':
    Utils_XML.PubMed_Doc2AuthorList(tree, fout)

  elif args.op == 'summary2abstract':
    if not ids: parser.error('--summary2abstract requires --ids.')
    n_out=0;
    for id_this in ids:
      logging.info(f"{id_this:08d}")
      if args.odir: fout = open(f"{args.odir}/{id_this:08d}_title_and_abstract.txt", "w")
      ok = Utils_XML.PubMed_Docsum2AbstractText(tree, id_this, fout, args.verbose)
      if ok: n_out+=1
      else: logging.error(f"No title or abstract found for PMID: {id_this}")
      if odir: fout.close()
    logging.info(f"Abstracts found: {n_out}")

  elif args.op == 'full2abstract':
    if not ids: parser.error('--full2abstract requires --ids.')
    n_out=0;
    for id_this in ids:
      if args.verbose:
        logging.info(f"{id_this:08d}")
      if args.odir: fout = open(f"{args.odir}/{id_this:08d}_title_and_abstract.txt", "w")
      ok = Utils_XML.PubMed_Doc2AbstractText(tree, id_this, fout, args.verbose)
      if ok: n_out+=1
      else: logging.error(f"No title or abstract found for PMID: {id_this}")
      if args.odir: fout.close()
    logging.info('abstracts found: %d'%n_out)

  else:
    parser.error(f"Operation invalid: {args.op}.")
